"""Core Library Utilities."""

from __future__ import annotations

import asyncio
import functools
import hashlib
import logging
import secrets
import socket
import ssl
import time
import urllib.parse
from collections.abc import Callable, Coroutine
from datetime import datetime, timedelta, timezone
from pathlib import Path
from types import TracebackType
from typing import Any, Self, TypeVar

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID

from pywebtransport.constants import (
    DEFAULT_LOG_FORMAT,
    DEFAULT_LOG_LEVEL,
    DEFAULT_PORT,
    DEFAULT_SECURE_PORT,
    MAX_STREAM_ID,
    SECURE_SCHEMES,
    WEBTRANSPORT_SCHEMES,
)
from pywebtransport.exceptions import CertificateError, ConfigurationError, TimeoutError, certificate_not_found
from pywebtransport.types import URL, Address, Buffer, Timeout, URLParts

__all__ = [
    "Timer",
    "build_webtransport_url",
    "calculate_checksum",
    "chunked_read",
    "create_task_with_timeout",
    "ensure_bytes",
    "ensure_str",
    "format_bytes",
    "format_duration",
    "format_timestamp",
    "generate_connection_id",
    "generate_request_id",
    "generate_self_signed_cert",
    "generate_session_id",
    "get_logger",
    "get_timestamp",
    "is_ipv4_address",
    "is_ipv6_address",
    "load_certificate",
    "merge_configs",
    "normalize_headers",
    "parse_webtransport_url",
    "resolve_address",
    "run_with_timeout",
    "setup_logging",
    "validate_address",
    "validate_error_code",
    "validate_port",
    "validate_session_id",
    "validate_stream_id",
    "validate_url",
    "wait_for_condition",
]

T = TypeVar("T")


class Timer:
    """A simple context manager for performance measurement."""

    def __init__(self, *, name: str = "timer") -> None:
        """Initialize the timer."""
        self.name = name
        self.start_time: float | None = None
        self.end_time: float | None = None

    @property
    def elapsed(self) -> float:
        """Get the elapsed time in seconds."""
        if self.start_time is None:
            return 0.0
        end = self.end_time or time.time()
        return end - self.start_time

    def __enter__(self) -> Self:
        """Start the timer upon entering the context."""
        self.start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Stop the timer and log the duration upon exiting the context."""
        elapsed = self.stop()
        logger = get_logger(name="timer")
        logger.debug("%s took %s", self.name, format_duration(seconds=elapsed))

    def start(self) -> None:
        """Start the timer."""
        self.start_time = time.time()
        self.end_time = None

    def stop(self) -> float:
        """Stop the timer and return the elapsed time."""
        if self.start_time is None:
            raise RuntimeError("Timer not started")
        self.end_time = time.time()
        return self.elapsed


def build_webtransport_url(
    *,
    host: str,
    port: int,
    path: str = "/",
    secure: bool = True,
    query_params: dict[str, str] | None = None,
) -> URL:
    """Build a WebTransport URL from its components."""
    scheme = "https" if secure else "http"
    if ":" in host and not host.startswith("["):
        host = f"[{host}]"
    query = ""
    if query_params:
        query = "?" + urllib.parse.urlencode(query_params)
    if (secure and port == DEFAULT_SECURE_PORT) or (not secure and port == DEFAULT_PORT):
        return f"{scheme}://{host}{path}{query}"
    return f"{scheme}://{host}:{port}{path}{query}"


def calculate_checksum(*, data: bytes, algorithm: str = "sha256") -> str:
    """Calculate the checksum of data using a specified secure hash algorithm."""
    if algorithm not in hashlib.algorithms_guaranteed:
        raise ValueError(f"Unsupported or insecure hash algorithm: {algorithm}")
    hash_obj = hashlib.new(algorithm)
    hash_obj.update(data)
    return hash_obj.hexdigest()


def chunked_read(*, data: bytes, chunk_size: int = 8192) -> list[bytes]:
    """Split data into a list of chunks of a specified size."""
    if chunk_size <= 0:
        raise ValueError("Chunk size must be positive")
    return [data[i : i + chunk_size] for i in range(0, len(data), chunk_size)]


async def create_task_with_timeout(
    *, coro: Coroutine[Any, Any, T], timeout: Timeout | None = None, name: str | None = None
) -> asyncio.Task[T]:
    """Create an asyncio task with an optional timeout wrapper."""
    if timeout is None:
        return asyncio.create_task(coro=coro, name=name)

    async def _wrapper() -> T:
        return await asyncio.wait_for(coro, timeout=timeout)

    return asyncio.create_task(coro=_wrapper(), name=name)


def ensure_bytes(*, data: Buffer | str, encoding: str = "utf-8") -> bytes:
    """Ensure that the given data is in bytes format."""
    match data:
        case str():
            return data.encode(encoding)
        case bytes():
            return data
        case bytearray() | memoryview():
            return bytes(data)
        case _:
            raise TypeError(f"Expected str, bytes, bytearray, or memoryview, got {type(data).__name__}")


def ensure_str(*, data: Buffer | str, encoding: str = "utf-8") -> str:
    """Ensure that the given data is in string format."""
    match data:
        case str():
            return data
        case bytes() | bytearray() | memoryview():
            return bytes(data).decode(encoding)
        case _:
            raise TypeError(f"Expected str, bytes, bytearray, or memoryview, got {type(data).__name__}")


def format_bytes(*, data: bytes, max_length: int = 100) -> str:
    """Format bytes for readable debugging output."""
    if len(data) <= max_length:
        return repr(data)
    return f"{repr(data[:max_length])}... ({len(data)} bytes total)"


def format_duration(*, seconds: float) -> str:
    """Format a duration in seconds into a human-readable string."""
    if seconds < 1:
        return f"{seconds * 1000:.1f}ms"
    if seconds < 60:
        return f"{seconds:.1f}s"
    if seconds < 3600:
        minutes = int(seconds // 60)
        secs = seconds % 60
        return f"{minutes}m{secs:.1f}s"
    hours = int(seconds // 3600)
    minutes = int((seconds % 3600) // 60)
    secs = seconds % 60
    return f"{hours}h{minutes}m{secs:.1f}s"


def format_timestamp(*, timestamp: float) -> str:
    """Format a Unix timestamp into an ISO 8601 string."""
    return datetime.fromtimestamp(timestamp).isoformat()


def generate_connection_id() -> str:
    """Generate a unique, URL-safe connection ID."""
    return secrets.token_urlsafe(12)


def generate_request_id() -> str:
    """Generate a unique, URL-safe request ID."""
    return secrets.token_urlsafe(8)


def generate_self_signed_cert(
    *, hostname: str, output_dir: str = ".", key_size: int = 2048, days_valid: int = 365
) -> tuple[str, str]:
    """Generate a self-signed certificate and key for testing purposes."""
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=key_size)
    subject = issuer = x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "CA"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, "San Francisco"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "pywebtransport"),
            x509.NameAttribute(NameOID.COMMON_NAME, hostname),
        ]
    )
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(timezone.utc))
        .not_valid_after(datetime.now(timezone.utc) + timedelta(days=days_valid))
        .add_extension(x509.SubjectAlternativeName([x509.DNSName(hostname)]), critical=False)
        .sign(private_key, hashes.SHA256())
    )
    output_path = Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)
    cert_file = output_path / f"{hostname}.crt"
    key_file = output_path / f"{hostname}.key"
    with open(cert_file, "wb") as f:
        f.write(cert.public_bytes(serialization.Encoding.PEM))
    with open(key_file, "wb") as f:
        f.write(
            private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )
    return (str(cert_file), str(key_file))


def generate_session_id() -> str:
    """Generate a unique, URL-safe session ID."""
    return secrets.token_urlsafe(16)


def get_logger(*, name: str) -> logging.Logger:
    """Get a logger instance with a specific name."""
    return logging.getLogger(f"pywebtransport.{name}")


def get_timestamp() -> float:
    """Get the current Unix timestamp."""
    return time.time()


@functools.cache
def is_ipv4_address(*, host: str) -> bool:
    """Check if a host string is a valid IPv4 address."""
    try:
        socket.inet_aton(host)
        return True
    except socket.error:
        return False


@functools.cache
def is_ipv6_address(*, host: str) -> bool:
    """Check if a host string is a valid IPv6 address."""
    try:
        socket.inet_pton(socket.AF_INET6, host)
        return True
    except socket.error:
        return False


def load_certificate(*, certfile: str, keyfile: str) -> ssl.SSLContext:
    """Load an SSL certificate and private key into an SSL context."""
    if not Path(certfile).exists():
        raise certificate_not_found(path=certfile)
    if not Path(keyfile).exists():
        raise certificate_not_found(path=keyfile)

    try:
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        context.minimum_version = ssl.TLSVersion.TLSv1_3
        context.load_cert_chain(certfile=certfile, keyfile=keyfile)
        return context
    except Exception as e:
        raise CertificateError(message=f"Failed to load certificate: {e}")


def merge_configs(*, base_config: dict[str, Any], override_config: dict[str, Any]) -> dict[str, Any]:
    """Recursively merge two configuration dictionaries."""
    result = base_config.copy()
    for key, value in override_config.items():
        match (result.get(key), value):
            case (dict() as base_dict, dict() as override_dict):
                result[key] = merge_configs(base_config=base_dict, override_config=override_dict)
            case _:
                result[key] = value
    return result


def normalize_headers(*, headers: dict[str, Any]) -> dict[str, str]:
    """Normalize header keys to lowercase and values to strings."""
    return {str(key).lower(): str(value) for key, value in headers.items()}


@functools.cache
def parse_webtransport_url(*, url: URL) -> URLParts:
    """Parse a WebTransport URL into its host, port, and path components."""
    parsed = urllib.parse.urlparse(url)
    if parsed.scheme not in WEBTRANSPORT_SCHEMES:
        raise ConfigurationError(
            message=f"Unsupported scheme '{parsed.scheme}'. Must be one of: {WEBTRANSPORT_SCHEMES}",
            config_key="url",
        )
    if not parsed.hostname:
        raise ConfigurationError(message="Missing hostname in URL", config_key="url")

    match parsed:
        case urllib.parse.ParseResult(port=p) if p:
            port = p
        case urllib.parse.ParseResult(scheme=s) if s in SECURE_SCHEMES:
            port = DEFAULT_SECURE_PORT
        case _:
            port = DEFAULT_PORT

    path = parsed.path or "/"
    if parsed.query:
        path += f"?{parsed.query}"
    if parsed.fragment:
        path += f"#{parsed.fragment}"
    return (parsed.hostname, port, path)


async def resolve_address(*, host: str, port: int, family: int = socket.AF_UNSPEC) -> Address:
    """Resolve a hostname to an IP address asynchronously."""
    try:
        loop = asyncio.get_running_loop()
        result = await loop.getaddrinfo(host=host, port=port, family=family, type=socket.SOCK_DGRAM)
        if not result:
            raise OSError(f"No address found for {host}:{port}")
        family, type_, proto, canonname, sockaddr = result[0]
        return (sockaddr[0], sockaddr[1])
    except OSError as e:
        raise ConfigurationError(message=f"Failed to resolve address {host}:{port}: {e}")


async def run_with_timeout(*, coro: Coroutine[Any, Any, T], timeout: float, default_value: T | None = None) -> T | None:
    """Run a coroutine with a timeout, returning a default value on timeout."""
    try:
        return await asyncio.wait_for(coro, timeout=timeout)
    except asyncio.TimeoutError:
        return default_value


def setup_logging(
    *,
    level: str = DEFAULT_LOG_LEVEL,
    format_string: str | None = None,
    logger_name: str = "pywebtransport",
) -> logging.Logger:
    """Set up and configure the main logger."""
    logger = logging.getLogger(logger_name)
    if logger.handlers:
        return logger
    numeric_level = getattr(logging, level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f"Invalid log level: {level}")
    logger.setLevel(numeric_level)
    handler = logging.StreamHandler()
    handler.setLevel(numeric_level)
    formatter = logging.Formatter(format_string or DEFAULT_LOG_FORMAT)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def validate_address(*, address: Any) -> None:
    """Validate a (host, port) address tuple."""
    if not isinstance(address, tuple) or len(address) != 2:
        raise TypeError("Address must be a (host, port) tuple")
    if not isinstance(address[0], str):
        raise TypeError("Host in address tuple must be a string")
    if not isinstance(address[1], int) or not (1 <= address[1] <= 65535):
        raise ValueError(f"Port must be an integer between 1 and 65535, got {address[1]}")


def validate_error_code(*, error_code: Any) -> None:
    """Validate a protocol error code."""
    if not isinstance(error_code, int):
        raise TypeError("Error code must be an integer")
    if not (0 <= error_code <= 2**32 - 1):
        raise ValueError(f"Error code {error_code} out of valid range")


def validate_port(*, port: Any) -> None:
    """Validate that a value is a valid network port."""
    if not isinstance(port, int) or not (1 <= port <= 65535):
        raise ValueError(f"Port must be an integer between 1 and 65535, got {port}")


def validate_session_id(*, session_id: Any) -> None:
    """Validate a WebTransport session ID."""
    if not isinstance(session_id, str):
        raise TypeError("Session ID must be a string")
    if not session_id:
        raise ValueError("Session ID cannot be empty")


def validate_stream_id(*, stream_id: Any) -> None:
    """Validate a WebTransport stream ID."""
    if not isinstance(stream_id, int):
        raise TypeError("Stream ID must be an integer")
    if not (0 <= stream_id <= MAX_STREAM_ID):
        raise ValueError(f"Stream ID {stream_id} out of valid range")


def validate_url(*, url: URL) -> bool:
    """Validate the format of a WebTransport URL."""
    try:
        parse_webtransport_url(url=url)
        return True
    except Exception:
        return False


async def wait_for_condition(
    *, condition: Callable[[], bool], timeout: Timeout | None = None, interval: float = 0.1
) -> None:
    """Wait for a condition to become true, with an optional timeout."""

    async def _waiter() -> None:
        while not condition():
            await asyncio.sleep(delay=interval)

    if timeout is None:
        await _waiter()
    else:
        try:
            await asyncio.wait_for(_waiter(), timeout=timeout)
        except asyncio.TimeoutError as e:
            raise TimeoutError(message=f"Condition not met within {timeout}s timeout") from e



"""WebTransport Server Implementation."""

from __future__ import annotations

import asyncio
import weakref
from dataclasses import asdict, dataclass
from pathlib import Path
from types import TracebackType
from typing import Any, Self, cast

from aioquic.asyncio import serve as quic_serve
from aioquic.asyncio.protocol import QuicConnectionProtocol
from aioquic.asyncio.server import QuicServer
from aioquic.quic.events import QuicEvent

from pywebtransport.config import ServerConfig
from pywebtransport.connection import ConnectionManager, WebTransportConnection
from pywebtransport.events import Event, EventEmitter
from pywebtransport.exceptions import ServerError
from pywebtransport.protocol.utils import create_quic_configuration
from pywebtransport.session import SessionManager
from pywebtransport.types import Address, EventType
from pywebtransport.utils import get_logger, get_timestamp

__all__ = [
    "ServerStats",
    "WebTransportServer",
    "WebTransportServerProtocol",
]

logger = get_logger(name="server.server")


@dataclass(kw_only=True)
class ServerStats:
    """A data class for storing server statistics."""

    connections_accepted: int = 0
    connections_rejected: int = 0
    connection_errors: int = 0
    protocol_errors: int = 0
    uptime: float = 0.0

    def to_dict(self) -> dict[str, Any]:
        """Convert statistics to a dictionary."""
        return asdict(obj=self)


class WebTransportServerProtocol(QuicConnectionProtocol):
    """The aioquic protocol implementation for the WebTransport server."""

    _server_ref: weakref.ReferenceType[WebTransportServer]
    _connection_ref: weakref.ReferenceType[WebTransportConnection] | None
    _event_queue: asyncio.Queue[QuicEvent]
    _event_processor_task: asyncio.Task[None] | None

    def __init__(self, server: WebTransportServer, *args: Any, **kwargs: Any) -> None:
        """Initialize the server protocol."""
        super().__init__(*args, **kwargs)
        self._server_ref = weakref.ref(server)
        self._connection_ref = None
        self._event_queue = asyncio.Queue()
        self._event_processor_task = None

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Handle a new connection being made."""
        super().connection_made(transport)
        self._event_processor_task = asyncio.create_task(self._process_events_loop())
        if server := self._server_ref():
            server._active_protocols.add(self)
            asyncio.create_task(server._handle_new_connection(transport=transport, protocol=self))

    def connection_lost(self, exc: Exception | None) -> None:
        """Handle a connection being lost."""
        super().connection_lost(exc)
        if self._event_processor_task and not self._event_processor_task.done():
            self._event_processor_task.cancel()
        if self._connection_ref and (connection := self._connection_ref()):
            connection._on_connection_lost(exc=exc)
        if server := self._server_ref():
            server._active_protocols.discard(self)

    def quic_event_received(self, event: QuicEvent) -> None:
        """Handle a QUIC event from the underlying transport."""
        self._event_queue.put_nowait(event)

    def set_connection(self, *, connection: WebTransportConnection) -> None:
        """Set a weak reference to the managing WebTransportConnection."""
        self._connection_ref = weakref.ref(connection)

    def transmit(self) -> None:
        """Transmit any pending data."""
        if self._transport is not None and not self._transport.is_closing():
            super().transmit()

    async def _process_events_loop(self) -> None:
        """Continuously process events from the QUIC connection."""
        try:
            conn: WebTransportConnection | None = None
            while True:
                event = await self._event_queue.get()

                if conn is None:
                    if self._connection_ref and (conn := self._connection_ref()):
                        pass
                    else:
                        await self._event_queue.put(event)
                        await asyncio.sleep(0.01)
                        continue

                if conn.protocol_handler:
                    try:
                        await conn.protocol_handler.handle_quic_event(event=event)
                    except Exception as e:
                        logger.error(
                            "Error handling QUIC event for %s: %s",
                            conn.connection_id,
                            e,
                            exc_info=True,
                        )
                else:
                    logger.warning(
                        "No handler available to process event for %s: %r",
                        conn.connection_id,
                        event,
                    )

        except asyncio.CancelledError:
            pass
        except Exception as e:
            logger.critical("Fatal error in server event processing loop: %s", e, exc_info=True)


class WebTransportServer(EventEmitter):
    """The main WebTransport server, managing lifecycle and connections."""

    def __init__(self, *, config: ServerConfig | None = None) -> None:
        """Initialize the WebTransport server."""
        super().__init__()
        self._config = config or ServerConfig.create()
        self._config.validate()
        self._serving, self._closing = False, False
        self._server: QuicServer | None = None
        self._start_time: float | None = None
        self._connection_manager = ConnectionManager.create(
            max_connections=self._config.max_connections,
            connection_cleanup_interval=self._config.connection_cleanup_interval,
            connection_idle_check_interval=self._config.connection_idle_check_interval,
            connection_idle_timeout=self._config.connection_idle_timeout,
        )
        self._session_manager = SessionManager.create(
            max_sessions=self._config.max_sessions,
            session_cleanup_interval=self._config.session_cleanup_interval,
        )
        self._background_tasks: list[asyncio.Task[Any]] = []
        self._active_protocols: set[WebTransportServerProtocol] = set()
        self._stats = ServerStats()
        self._setup_event_handlers()
        logger.info("WebTransport server initialized.")

    @property
    def is_serving(self) -> bool:
        """Check if the server is currently serving."""
        return self._serving

    @property
    def config(self) -> ServerConfig:
        """Get the server's configuration object."""
        return self._config

    @property
    def local_address(self) -> Address | None:
        """Get the local address the server is bound to."""
        if self._server and hasattr(self._server, "_transport") and self._server._transport:
            return cast(Address | None, self._server._transport.get_extra_info("sockname"))
        return None

    @property
    def session_manager(self) -> SessionManager:
        """Get the server's session manager instance."""
        return self._session_manager

    async def __aenter__(self) -> Self:
        """Enter the async context for the server."""
        await self._connection_manager.__aenter__()
        await self._session_manager.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the async context and close the server."""
        await self.close()

    async def close(self) -> None:
        """Gracefully shut down the server and its resources."""
        if not self._serving or self._closing:
            return
        logger.info("Closing WebTransport server...")
        self._closing = True
        for task in self._background_tasks:
            if not task.done():
                task.cancel()
        await asyncio.gather(*self._background_tasks, return_exceptions=True)

        try:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(self._connection_manager.shutdown())
                tg.create_task(self._session_manager.shutdown())
        except* Exception as eg:
            logger.error(
                "Errors occurred during manager shutdown: %s",
                eg.exceptions,
                exc_info=eg,
            )

        if self._server:
            self._server.close()
            if hasattr(self._server, "wait_closed"):
                await self._server.wait_closed()
        self._serving, self._closing = False, False
        logger.info("WebTransport server closed.")

    async def listen(self, *, host: str | None = None, port: int | None = None) -> None:
        """Start the server and begin listening for connections."""
        if self._serving:
            raise ServerError(message="Server is already serving")

        bind_host, bind_port = host or self._config.bind_host, port or self._config.bind_port
        logger.info("Starting WebTransport server on %s:%s", bind_host, bind_port)
        try:
            quic_config = create_quic_configuration(is_client=False, **self._config.to_dict())
            quic_config.load_cert_chain(certfile=Path(self._config.certfile), keyfile=Path(self._config.keyfile))
            if self._config.ca_certs:
                quic_config.load_verify_locations(cafile=self._config.ca_certs)
            quic_config.verify_mode = self._config.verify_mode
            self._server = await quic_serve(
                host=bind_host,
                port=bind_port,
                configuration=quic_config,
                create_protocol=lambda *a, **kw: WebTransportServerProtocol(self, *a, **kw),
            )
            self._serving, self._start_time = True, get_timestamp()
            self._start_background_tasks()
            logger.info("WebTransport server listening on %s", self.local_address)
        except Exception as e:
            logger.critical("Failed to start server: %s", e, exc_info=True)
            raise ServerError(message=f"Failed to start server: {e}") from e

    async def serve_forever(self) -> None:
        """Run the server indefinitely until interrupted."""
        if not self._serving or not self._server:
            raise ServerError(message="Server is not listening")

        logger.info("Server is running. Press Ctrl+C to stop.")
        try:
            if hasattr(self._server, "wait_closed"):
                await self._server.wait_closed()
            else:
                while self._serving:
                    await asyncio.sleep(3600)
        except (asyncio.CancelledError, KeyboardInterrupt):
            logger.info("Server stop signal received.")
        finally:
            await self.close()

    async def debug_state(self) -> dict[str, Any]:
        """Get a detailed snapshot of the server's state for debugging."""
        stats, connections, sessions = await asyncio.gather(
            self.get_server_stats(),
            self._connection_manager.get_all_connections(),
            self._session_manager.get_all_sessions(),
        )

        session_stats_tasks = [sess.get_session_stats() for sess in sessions]
        session_stats_list = await asyncio.gather(*session_stats_tasks)

        return {
            "server_info": {
                "serving": self.is_serving,
                "local_address": self.local_address,
            },
            "aggregated_stats": stats,
            "connections": [conn.info.to_dict() for conn in connections],
            "sessions": session_stats_list,
        }

    async def diagnose_issues(self) -> list[str]:
        """Analyze server statistics and configuration to identify potential issues."""
        issues: list[str] = []
        stats = await self.get_server_stats()

        if not self.is_serving:
            issues.append("Server is not currently serving.")

        connections_stats = stats.get("connections", {})
        accepted = stats.get("connections_accepted", 0)
        rejected = stats.get("connections_rejected", 0)
        total_conn_attempts = accepted + rejected
        if total_conn_attempts > 20 and total_conn_attempts > 0 and (rejected / total_conn_attempts) > 0.1:
            issues.append(f"High connection rejection rate: {rejected}/{total_conn_attempts}")

        max_connections = self.config.max_connections
        active_connections = connections_stats.get("active", 0) if connections_stats else 0
        if max_connections > 0 and (active_connections / max_connections) > 0.9:
            issues.append(f"High connection usage: {active_connections / max_connections:.1%}")

        try:
            if not Path(self.config.certfile).exists():
                issues.append(f"Certificate file not found: {self.config.certfile}")
            if not Path(self.config.keyfile).exists():
                issues.append(f"Key file not found: {self.config.keyfile}")
        except Exception:
            issues.append("Certificate configuration appears invalid.")

        return issues

    async def get_server_stats(self) -> dict[str, Any]:
        """Get a snapshot of the server's performance statistics."""
        if self._start_time:
            self._stats.uptime = get_timestamp() - self._start_time

        base_stats = self._stats.to_dict()

        conn_stats, sess_stats = await asyncio.gather(
            self._connection_manager.get_stats(),
            self._session_manager.get_stats(),
        )

        base_stats["connections"] = conn_stats
        base_stats["sessions"] = sess_stats
        return base_stats

    async def _handle_new_connection(
        self, *, transport: asyncio.BaseTransport, protocol: WebTransportServerProtocol
    ) -> None:
        """Handle a new incoming connection and set up the event forwarding chain."""
        connection: WebTransportConnection | None = None
        try:
            connection = WebTransportConnection(config=self._config)

            server_ref = weakref.ref(self)
            conn_ref = weakref.ref(connection)

            async def forward_session_request(event: Event) -> None:
                server = server_ref()
                conn = conn_ref()
                if server and conn and isinstance(event.data, dict):
                    event_data = event.data.copy()
                    if "connection" not in event_data:
                        event_data["connection"] = conn
                    logger.debug(
                        "Forwarding session request for path '%s' from connection %s to server.",
                        event_data.get("path"),
                        conn.connection_id,
                    )
                    await server.emit(event_type=EventType.SESSION_REQUEST, data=event_data)

            connection.on(event_type=EventType.SESSION_REQUEST, handler=forward_session_request)
            dgram_transport = cast(asyncio.DatagramTransport, transport)
            await connection.accept(transport=dgram_transport, protocol=protocol)
            await self._connection_manager.add_connection(connection=connection)
            self._stats.connections_accepted += 1
            logger.info("New connection accepted: %s", connection.connection_id)
        except Exception as e:
            self._stats.connections_rejected += 1
            self._stats.connection_errors += 1
            logger.error("Failed to handle new connection: %s", e, exc_info=True)
            if connection and not connection.is_closed:
                await connection.close()
            try:
                transport.close()
            except Exception:
                pass

    def _setup_event_handlers(self) -> None:
        """Set up internal event handlers."""
        pass

    def _start_background_tasks(self) -> None:
        """Start all background tasks for the server."""
        pass

    def __str__(self) -> str:
        """Format a concise summary of server information for logging."""
        status = "serving" if self.is_serving else "stopped"
        address = self.local_address or ("unknown", 0)
        conn_count = self._connection_manager.get_connection_count()
        sess_count = self._session_manager.get_session_count()
        return (
            f"WebTransportServer(status={status}, "
            f"address={address[0]}:{address[1]}, "
            f"connections={conn_count}, "
            f"sessions={sess_count})"
        )



"""Core Type Definitions."""

from __future__ import annotations

import ssl
from collections.abc import AsyncGenerator, AsyncIterator, Awaitable, Callable
from enum import StrEnum
from typing import TYPE_CHECKING, Any, AsyncContextManager, Protocol, TypeAlias, TypeVar, runtime_checkable

if TYPE_CHECKING:
    from pywebtransport.connection import WebTransportConnection
    from pywebtransport.events import Event
    from pywebtransport.session import WebTransportSession
    from pywebtransport.stream import WebTransportReceiveStream, WebTransportStream

__all__ = [
    "Address",
    "AsyncContextManager",
    "AsyncGenerator",
    "AsyncIterator",
    "AuthHandlerProtocol",
    "BidirectionalStreamProtocol",
    "Buffer",
    "BufferSize",
    "CertificateData",
    "ClientConfigProtocol",
    "ConnectionId",
    "ConnectionInfoProtocol",
    "ConnectionLostHandler",
    "ConnectionState",
    "ConnectionStats",
    "Data",
    "DatagramHandler",
    "ErrorCode",
    "ErrorHandler",
    "EventData",
    "EventHandler",
    "EventEmitterProtocol",
    "EventType",
    "FlowControlWindow",
    "Headers",
    "MiddlewareProtocol",
    "Priority",
    "PrivateKeyData",
    "ReadableStreamProtocol",
    "ReasonPhrase",
    "RouteHandler",
    "RoutePattern",
    "Routes",
    "SSLContext",
    "Serializer",
    "ServerConfigProtocol",
    "SessionHandler",
    "SessionId",
    "SessionInfoProtocol",
    "SessionState",
    "SessionStats",
    "StreamDirection",
    "StreamHandler",
    "StreamId",
    "StreamInfoProtocol",
    "StreamState",
    "StreamStats",
    "Timestamp",
    "Timeout",
    "TimeoutDict",
    "URL",
    "URLParts",
    "WebTransportProtocol",
    "Weight",
    "WritableStreamProtocol",
]

T = TypeVar("T")
P = TypeVar("P")


class ConnectionState(StrEnum):
    """Enumeration of connection states."""

    IDLE = "idle"
    CONNECTING = "connecting"
    CONNECTED = "connected"
    CLOSING = "closing"
    CLOSED = "closed"
    FAILED = "failed"
    DRAINING = "draining"


class EventType(StrEnum):
    """Enumeration of system event types."""

    CAPSULE_RECEIVED = "capsule_received"
    CONNECTION_ESTABLISHED = "connection_established"
    CONNECTION_LOST = "connection_lost"
    CONNECTION_FAILED = "connection_failed"
    CONNECTION_CLOSED = "connection_closed"
    DATAGRAM_ERROR = "datagram_error"
    DATAGRAM_RECEIVED = "datagram_received"
    DATAGRAM_SENT = "datagram_sent"
    PROTOCOL_ERROR = "protocol_error"
    SETTINGS_RECEIVED = "settings_received"
    SESSION_CLOSED = "session_closed"
    SESSION_DRAINING = "session_draining"
    SESSION_MAX_DATA_UPDATED = "session_max_data_updated"
    SESSION_MAX_STREAMS_BIDI_UPDATED = "session_max_streams_bidi_updated"
    SESSION_MAX_STREAMS_UNI_UPDATED = "session_max_streams_uni_updated"
    SESSION_READY = "session_ready"
    SESSION_REQUEST = "session_request"
    STREAM_CLOSED = "stream_closed"
    STREAM_DATA_RECEIVED = "stream_data_received"
    STREAM_ERROR = "stream_error"
    STREAM_OPENED = "stream_opened"
    TIMEOUT_ERROR = "timeout_error"


class SessionState(StrEnum):
    """Enumeration of WebTransport session states."""

    CONNECTING = "connecting"
    CONNECTED = "connected"
    CLOSING = "closing"
    DRAINING = "draining"
    CLOSED = "closed"


class StreamDirection(StrEnum):
    """Enumeration of stream directions."""

    BIDIRECTIONAL = "bidirectional"
    SEND_ONLY = "send_only"
    RECEIVE_ONLY = "receive_only"


class StreamState(StrEnum):
    """Enumeration of WebTransport stream states."""

    IDLE = "idle"
    OPEN = "open"
    HALF_CLOSED_LOCAL = "half_closed_local"
    HALF_CLOSED_REMOTE = "half_closed_remote"
    CLOSED = "closed"
    RESET_SENT = "reset_sent"
    RESET_RECEIVED = "reset_received"


Address: TypeAlias = tuple[str, int]
Buffer: TypeAlias = bytes | bytearray | memoryview
BufferSize: TypeAlias = int
CertificateData: TypeAlias = str | bytes
ConnectionId: TypeAlias = str
ConnectionStats: TypeAlias = dict[str, int | float | str | list["SessionStats"]]
Data: TypeAlias = bytes | str
ErrorCode: TypeAlias = int
EventData: TypeAlias = Any
FlowControlWindow: TypeAlias = int
Headers: TypeAlias = dict[str, str]
Priority: TypeAlias = int
PrivateKeyData: TypeAlias = str | bytes
ReasonPhrase: TypeAlias = str
RoutePattern: TypeAlias = str
SSLContext: TypeAlias = ssl.SSLContext
SessionId: TypeAlias = str
SessionStats: TypeAlias = dict[str, int | float | str | list["StreamStats"]]
StreamId: TypeAlias = int
StreamStats: TypeAlias = dict[str, int | float | str]
Timestamp: TypeAlias = float
Timeout: TypeAlias = float | None
TimeoutDict: TypeAlias = dict[str, float]
URL: TypeAlias = str
URLParts: TypeAlias = tuple[str, int, str]
Weight: TypeAlias = int


if TYPE_CHECKING:
    ConnectionLostHandler: TypeAlias = Callable[[WebTransportConnection, Exception | None], Awaitable[None]]
    EventHandler: TypeAlias = Callable[[Event], Awaitable[None]]
    RouteHandler: TypeAlias = Callable[[WebTransportSession], Awaitable[None]]
    SessionHandler: TypeAlias = Callable[[WebTransportSession], Awaitable[None]]
    StreamHandler: TypeAlias = Callable[[WebTransportStream | WebTransportReceiveStream], Awaitable[None]]
else:
    ConnectionLostHandler: TypeAlias = Callable[[Any, Exception | None], Awaitable[None]]
    EventHandler: TypeAlias = Callable[[Any], Awaitable[None]]
    RouteHandler: TypeAlias = Callable[[Any], Awaitable[None]]
    SessionHandler: TypeAlias = Callable[[Any], Awaitable[None]]
    StreamHandler: TypeAlias = Callable[[Any], Awaitable[None]]

DatagramHandler: TypeAlias = Callable[[bytes], Awaitable[None]]
ErrorHandler: TypeAlias = Callable[[Exception], Awaitable[None]]
Routes: TypeAlias = dict[RoutePattern, RouteHandler]


@runtime_checkable
class ClientConfigProtocol(Protocol):
    """A protocol defining the structure of a client configuration object."""

    alpn_protocols: list[str]
    auto_reconnect: bool
    ca_certs: str | None
    certfile: str | None
    close_timeout: float
    congestion_control_algorithm: str
    connect_timeout: float
    connection_cleanup_interval: float
    connection_idle_check_interval: float
    connection_idle_timeout: float
    connection_keepalive_timeout: float
    debug: bool
    flow_control_window_auto_scale: bool
    flow_control_window_size: int
    headers: Headers
    initial_max_data: int
    initial_max_streams_bidi: int
    initial_max_streams_uni: int
    keep_alive: bool
    keyfile: str | None
    log_level: str
    max_connections: int
    max_datagram_size: int
    max_incoming_streams: int
    max_pending_events_per_session: int
    max_retries: int
    max_retry_delay: float
    max_stream_buffer_size: int
    max_streams: int
    max_total_pending_events: int
    pending_event_ttl: float
    read_timeout: float | None
    retry_backoff: float
    retry_delay: float
    stream_buffer_size: int
    stream_cleanup_interval: float
    stream_creation_timeout: float
    stream_flow_control_increment_bidi: int
    stream_flow_control_increment_uni: int
    user_agent: str
    verify_mode: ssl.VerifyMode | None
    write_timeout: float | None


@runtime_checkable
class AuthHandlerProtocol(Protocol):
    """A protocol for auth handlers."""

    async def __call__(self, *, headers: Headers) -> bool: ...


@runtime_checkable
class ConnectionInfoProtocol(Protocol):
    """A protocol for retrieving connection information."""

    local_address: Address | None
    remote_address: Address | None
    state: ConnectionState
    established_at: float | None
    bytes_sent: int
    bytes_received: int
    streams_created: int
    datagrams_sent: int
    datagrams_received: int


@runtime_checkable
class EventEmitterProtocol(Protocol):
    """A protocol for an event emitter."""

    async def emit(self, *, event_type: EventType, data: EventData | None = None) -> None:
        """Emit an event."""
        ...

    def off(self, *, event_type: EventType, handler: EventHandler | None = None) -> None:
        """Unregister an event handler."""
        ...

    def on(self, *, event_type: EventType, handler: EventHandler) -> None:
        """Register an event handler."""
        ...


@runtime_checkable
class MiddlewareProtocol(Protocol):
    """A protocol for a middleware object."""

    async def __call__(self, *, session: WebTransportSession) -> bool: ...


@runtime_checkable
class ReadableStreamProtocol(Protocol):
    """A protocol for a readable stream."""

    def at_eof(self) -> bool:
        """Check if the end of the stream has been reached."""
        ...

    async def read(self, *, size: int = -1) -> bytes:
        """Read data from the stream."""
        ...

    async def readline(self, *, separator: bytes = b"\n") -> bytes:
        """Read a line from the stream."""
        ...

    async def readexactly(self, *, n: int) -> bytes:
        """Read exactly n bytes from the stream."""
        ...

    async def readuntil(self, *, separator: bytes = b"\n") -> bytes:
        """Read from the stream until a separator is found."""
        ...


@runtime_checkable
class Serializer(Protocol):
    """A protocol for serializing and deserializing structured data."""

    def serialize(self, *, obj: Any) -> bytes:
        """Serialize an object into bytes."""
        ...

    def deserialize(self, *, data: bytes, obj_type: Any = None) -> Any:
        """Deserialize bytes into an object."""
        ...


@runtime_checkable
class WritableStreamProtocol(Protocol):
    """A protocol for a writable stream."""

    async def close(self, *, code: int | None = None, reason: str | None = None) -> None:
        """Close the stream."""
        ...

    async def flush(self) -> None:
        """Flush the stream's write buffer."""
        ...

    def is_closing(self) -> bool:
        """Check if the stream is in the process of closing."""
        ...

    async def write(self, *, data: Data) -> None:
        """Write data to the stream."""
        ...

    async def writelines(self, *, lines: list[Data]) -> None:
        """Write multiple lines to the stream."""
        ...


@runtime_checkable
class BidirectionalStreamProtocol(ReadableStreamProtocol, WritableStreamProtocol, Protocol):
    """A protocol for a bidirectional stream."""

    pass


@runtime_checkable
class ServerConfigProtocol(Protocol):
    """A protocol defining the structure of a server configuration object."""

    access_log: bool
    alpn_protocols: list[str]
    bind_host: str
    bind_port: int
    ca_certs: str | None
    certfile: str
    congestion_control_algorithm: str
    connection_cleanup_interval: float
    connection_idle_check_interval: float
    connection_idle_timeout: float
    connection_keepalive_timeout: float
    debug: bool
    flow_control_window_auto_scale: bool
    flow_control_window_size: int
    initial_max_data: int
    initial_max_streams_bidi: int
    initial_max_streams_uni: int
    keep_alive: bool
    keyfile: str
    log_level: str
    max_connections: int
    max_datagram_size: int
    max_incoming_streams: int
    max_pending_events_per_session: int
    max_sessions: int
    max_stream_buffer_size: int
    max_streams_per_connection: int
    max_total_pending_events: int
    middleware: list[Any]
    pending_event_ttl: float
    read_timeout: float | None
    session_cleanup_interval: float
    stream_buffer_size: int
    stream_cleanup_interval: float
    stream_flow_control_increment_bidi: int
    stream_flow_control_increment_uni: int
    verify_mode: ssl.VerifyMode
    write_timeout: float | None


@runtime_checkable
class SessionInfoProtocol(Protocol):
    """A protocol for retrieving session information."""

    session_id: SessionId
    state: SessionState
    created_at: float
    ready_at: float | None
    closed_at: float | None
    streams_count: int
    bytes_sent: int
    bytes_received: int


@runtime_checkable
class StreamInfoProtocol(Protocol):
    """A protocol for retrieving stream information."""

    stream_id: StreamId
    direction: StreamDirection
    state: StreamState
    created_at: float
    closed_at: float | None
    bytes_sent: int
    bytes_received: int


@runtime_checkable
class WebTransportProtocol(Protocol):
    """A protocol for the underlying WebTransport transport layer."""

    def connection_made(self, transport: Any) -> None:
        """Called when a connection is established."""
        ...

    def connection_lost(self, exc: Exception | None) -> None:
        """Called when a connection is lost."""
        ...

    def datagram_received(self, data: bytes, addr: Address) -> None:
        """Called when a datagram is received."""
        ...

    def error_received(self, exc: Exception) -> None:
        """Called when an error is received."""
        ...



"""Built-in Pub/Sub framework for WebTransport."""

from .exceptions import NotSubscribedError, PubSubError, SubscriptionFailedError
from .manager import PubSubManager, PubSubStats, Subscription

__all__ = [
    "NotSubscribedError",
    "PubSubError",
    "PubSubManager",
    "PubSubStats",
    "Subscription",
    "SubscriptionFailedError",
]



"""WebTransport stream utility functions."""

from __future__ import annotations

from pywebtransport.exceptions import StreamError
from pywebtransport.stream.stream import WebTransportReceiveStream, WebTransportSendStream, WebTransportStream
from pywebtransport.utils import get_logger

__all__ = [
    "copy_stream_data",
    "echo_stream",
]

logger = get_logger(name="stream.utils")


async def copy_stream_data(
    *,
    source: WebTransportReceiveStream,
    destination: WebTransportSendStream,
    chunk_size: int = 8192,
) -> int:
    """Copy all data from a source stream to a destination stream."""
    total_bytes = 0
    try:
        async for chunk in source.read_iter(chunk_size=chunk_size):
            await destination.write(data=chunk)
            total_bytes += len(chunk)
        await destination.close()
    except StreamError as e:
        logger.error("Error copying stream data: %s", e, exc_info=True)
        await destination.abort(code=1)
        raise
    return total_bytes


async def echo_stream(*, stream: WebTransportStream) -> None:
    """Echo all data received on a bidirectional stream back to the sender."""
    try:
        async for chunk in stream.read_iter():
            await stream.write(data=chunk)
        await stream.close()
    except StreamError as e:
        logger.error("Error in echo stream: %s", e, exc_info=True)
        await stream.abort(code=1)



"""High-level Pub/Sub manager for WebTransport sessions."""

from __future__ import annotations

import asyncio
from collections import defaultdict
from dataclasses import dataclass, field
from types import TracebackType
from typing import TYPE_CHECKING, AsyncIterator, Self

from pywebtransport.constants import DEFAULT_PUBSUB_SUBSCRIPTION_QUEUE_SIZE
from pywebtransport.exceptions import ConnectionError, StreamError
from pywebtransport.pubsub.exceptions import NotSubscribedError, PubSubError, SubscriptionFailedError
from pywebtransport.stream import WebTransportStream
from pywebtransport.types import Data
from pywebtransport.utils import ensure_bytes, get_logger, get_timestamp

if TYPE_CHECKING:
    from pywebtransport.session import WebTransportSession


__all__ = ["PubSubManager", "PubSubStats", "Subscription"]

logger = get_logger(name="pubsub.manager")


@dataclass(kw_only=True)
class PubSubStats:
    """Represents statistics for the Pub/Sub manager."""

    created_at: float = field(default_factory=get_timestamp)
    topics_subscribed: int = 0
    messages_published: int = 0
    messages_received: int = 0
    subscription_errors: int = 0

    def to_dict(self) -> dict[str, float | int]:
        """Convert statistics to a dictionary."""
        return {
            "created_at": self.created_at,
            "topics_subscribed": self.topics_subscribed,
            "messages_published": self.messages_published,
            "messages_received": self.messages_received,
            "subscription_errors": self.subscription_errors,
        }


class Subscription:
    """Represents a subscription to a single topic."""

    def __init__(self, *, topic: str, manager: PubSubManager, max_queue_size: int) -> None:
        """Initialize the Subscription."""
        self._topic = topic
        self._manager = manager
        self._max_queue_size = max_queue_size
        self._queue: asyncio.Queue[bytes | None] | None = None

    async def __aenter__(self) -> Self:
        """Enter the async context and initialize the message queue."""
        if self._queue is None:
            self._queue = asyncio.Queue(maxsize=self._max_queue_size)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the async context and unsubscribe."""
        await self.unsubscribe()

    async def unsubscribe(self) -> None:
        """Unsubscribe from the topic."""
        await self._manager.unsubscribe(topic=self._topic)

    async def __aiter__(self) -> AsyncIterator[bytes]:
        """Iterate over messages from the topic."""
        if self._queue is None:
            raise PubSubError(
                message=(
                    "Subscription has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        while True:
            try:
                message = await self._queue.get()
                if message is None:
                    break
                yield message
            except asyncio.CancelledError:
                break


class PubSubManager:
    """Manages the Pub/Sub lifecycle over a single WebTransport session."""

    def __init__(self, *, session: WebTransportSession) -> None:
        """Initialize the PubSubManager."""
        self._session = session
        self._stream: WebTransportStream | None = None
        self._ingress_task: asyncio.Task[None] | None = None
        self._lock: asyncio.Lock | None = None
        self._subscriptions: dict[str, set[Subscription]] = defaultdict(set)
        self._pending_subscriptions: dict[str, tuple[asyncio.Future[Subscription], Subscription]] = {}
        self._stats = PubSubStats()
        self._is_closing = False

    @property
    def stats(self) -> PubSubStats:
        """Get the current statistics for the Pub/Sub manager."""
        return self._stats

    async def __aenter__(self) -> Self:
        """Enter the async context, ensuring the manager is initialized."""
        if self._lock is None:
            self._lock = asyncio.Lock()
        await self._ensure_initialized()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the async context, closing the manager."""
        await self.close()

    async def close(self) -> None:
        """Close the Pub/Sub manager and its underlying resources."""
        if self._is_closing:
            return

        self._is_closing = True

        if self._ingress_task and not self._ingress_task.done():
            self._ingress_task.cancel()
        if self._stream and not self._stream.is_closed:
            await self._stream.abort()

        for future, _ in self._pending_subscriptions.values():
            if not future.done():
                future.set_exception(PubSubError(message="PubSubManager is closing."))
        for subs in self._subscriptions.values():
            for sub in subs:
                if sub._queue and not sub._queue.full():
                    sub._queue.put_nowait(None)

        self._pending_subscriptions.clear()
        self._subscriptions.clear()

    async def publish(self, *, topic: str, data: Data) -> None:
        """Publish a message to a topic."""
        if self._lock is None:
            raise PubSubError(
                message=(
                    "PubSubManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            await self._ensure_initialized()

            if self._stream is None or self._stream.is_closed:
                raise PubSubError(message="Pub/Sub stream is not available.")

            data_bytes = ensure_bytes(data=data)
            command = b"PUB %b %d\n" % (topic.encode(), len(data_bytes))

            await self._stream.write(data=command + data_bytes)
            self._stats.messages_published += 1

    async def subscribe(
        self, *, topic: str, max_queue_size: int = DEFAULT_PUBSUB_SUBSCRIPTION_QUEUE_SIZE, timeout: float = 10.0
    ) -> Subscription:
        """Subscribe to a topic and return a subscription object."""
        if self._lock is None:
            raise PubSubError(
                message=(
                    "PubSubManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )
        if max_queue_size <= 0:
            raise ValueError("max_queue_size must be positive.")

        async with self._lock:
            await self._ensure_initialized()
            if topic in self._pending_subscriptions:
                raise PubSubError(message=f"Subscription to topic '{topic}' is already pending.")

            future: asyncio.Future[Subscription] = asyncio.get_running_loop().create_future()
            subscription = Subscription(topic=topic, manager=self, max_queue_size=max_queue_size)
            self._pending_subscriptions[topic] = (future, subscription)

        if self._stream is None or self._stream.is_closed:
            raise PubSubError(message="Pub/Sub stream is not available.")
        await self._stream.write(data=b"SUB %b\n" % topic.encode())

        try:
            confirmed_subscription = await asyncio.wait_for(future, timeout=timeout)
            self._subscriptions[topic].add(confirmed_subscription)
            self._stats.topics_subscribed = len(self._subscriptions)
            return confirmed_subscription
        except asyncio.TimeoutError:
            self._stats.subscription_errors += 1
            self._pending_subscriptions.pop(topic, None)
            raise SubscriptionFailedError(message=f"Subscription to topic '{topic}' timed out.") from None
        except Exception:
            self._stats.subscription_errors += 1
            self._pending_subscriptions.pop(topic, None)
            raise

    async def unsubscribe(self, *, topic: str) -> None:
        """Unsubscribe from a topic."""
        if self._lock is None:
            raise PubSubError(
                message=(
                    "PubSubManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            if not self._subscriptions.get(topic):
                return

            if self._stream is None or self._stream.is_closed:
                raise NotSubscribedError(message="Pub/Sub stream is not available to unsubscribe.")
            await self._stream.write(data=b"UNSUB %b\n" % topic.encode())

            for sub in self._subscriptions.get(topic, set()):
                if sub._queue and not sub._queue.full():
                    sub._queue.put_nowait(None)

            self._subscriptions.pop(topic, None)
            self._stats.topics_subscribed = len(self._subscriptions)

    async def _ensure_initialized(self) -> None:
        """Lazily create the Pub/Sub stream and start the reader task."""
        if self._stream is None:
            self._stream = await self._session.create_bidirectional_stream()
            self._ingress_task = asyncio.create_task(self._ingress_loop())
            self._ingress_task.add_done_callback(self._on_ingress_done)

    async def _ingress_loop(self) -> None:
        """Read and dispatch incoming messages and signals."""
        if self._stream is None:
            return

        try:
            while not self._stream.is_closed:
                line = await self._stream.readline()
                if not line:
                    break
                parts = line.strip().split(b" ", 2)
                command = parts[0]

                if command == b"MSG" and len(parts) == 3:
                    topic, length_str = parts[1].decode(), parts[2]
                    length = int(length_str)
                    payload = await self._stream.readexactly(n=length)
                    if subs := self._subscriptions.get(topic):
                        self._stats.messages_received += 1
                        for sub in subs.copy():
                            if sub._queue:
                                try:
                                    sub._queue.put_nowait(payload)
                                except asyncio.QueueFull:
                                    logger.warning(
                                        "Subscription queue for topic '%s' is full. Dropping message.", topic
                                    )
                elif command == b"SUB-OK" and len(parts) == 2:
                    topic = parts[1].decode()
                    if pending := self._pending_subscriptions.pop(topic, None):
                        future, subscription = pending
                        if not future.done():
                            future.set_result(subscription)
                elif command == b"SUB-FAIL" and len(parts) == 3:
                    topic, reason = parts[1].decode(), parts[2].decode()
                    if pending := self._pending_subscriptions.pop(topic, None):
                        future, _ = pending
                        if not future.done():
                            future.set_exception(SubscriptionFailedError(reason))
        except (ConnectionError, StreamError, asyncio.IncompleteReadError) as e:
            logger.info("Pub/Sub stream closed: %s", e)
            raise
        except Exception as e:
            logger.error("Error in Pub/Sub ingress loop: %s", e, exc_info=True)
            raise

    def _on_ingress_done(self, task: asyncio.Task[None]) -> None:
        """Callback to ensure cleanup is triggered when the ingress task finishes."""
        if self._is_closing:
            return

        if not task.cancelled():
            if exc := task.exception():
                logger.error(
                    "Pub/Sub ingress task finished unexpectedly with an exception: %s.",
                    exc,
                    exc_info=exc,
                )

        asyncio.create_task(self.close())



"""WebTransport Connection Subpackage."""

from .connection import ConnectionInfo, WebTransportConnection
from .load_balancer import ConnectionLoadBalancer
from .manager import ConnectionManager
from .pool import ConnectionPool

__all__ = [
    "ConnectionInfo",
    "ConnectionLoadBalancer",
    "ConnectionManager",
    "ConnectionPool",
    "WebTransportConnection",
]



"""WebTransport Protocol Utility Functions."""

from __future__ import annotations

from typing import Any

from aioquic.quic.configuration import QuicConfiguration

from pywebtransport.constants import ErrorCodes
from pywebtransport.types import ConnectionState, SessionState, StreamDirection, StreamId
from pywebtransport.utils import get_logger, validate_stream_id

__all__ = [
    "can_receive_data",
    "can_receive_data_on_stream",
    "can_send_data",
    "can_send_data_on_stream",
    "create_quic_configuration",
    "get_stream_direction_from_id",
    "is_bidirectional_stream",
    "is_client_initiated_stream",
    "is_server_initiated_stream",
    "is_unidirectional_stream",
    "webtransport_code_to_http_code",
]

logger = get_logger(name="protocol.utils")


def can_receive_data(*, connection_state: ConnectionState, session_state: SessionState) -> bool:
    """Check if data can be received based on connection and session states."""
    return connection_state == ConnectionState.CONNECTED and session_state in [
        SessionState.CONNECTED,
        SessionState.DRAINING,
    ]


def can_receive_data_on_stream(*, stream_id: StreamId, is_client: bool) -> bool:
    """Check if the local endpoint can receive data on a given stream."""
    if is_bidirectional_stream(stream_id=stream_id):
        return True
    return (is_client and is_server_initiated_stream(stream_id=stream_id)) or (
        not is_client and is_client_initiated_stream(stream_id=stream_id)
    )


def can_send_data(*, connection_state: ConnectionState, session_state: SessionState) -> bool:
    """Check if data can be sent based on connection and session states."""
    return connection_state == ConnectionState.CONNECTED and session_state == SessionState.CONNECTED


def can_send_data_on_stream(*, stream_id: StreamId, is_client: bool) -> bool:
    """Check if the local endpoint can send data on a given stream."""
    if is_bidirectional_stream(stream_id=stream_id):
        return True
    return (is_client and is_client_initiated_stream(stream_id=stream_id)) or (
        not is_client and is_server_initiated_stream(stream_id=stream_id)
    )


def create_quic_configuration(*, is_client: bool = True, **kwargs: Any) -> QuicConfiguration:
    """Create a QUIC configuration from keyword arguments."""
    config_params = {
        "alpn_protocols": kwargs["alpn_protocols"],
        "congestion_control_algorithm": kwargs["congestion_control_algorithm"],
        "max_datagram_frame_size": kwargs["max_datagram_size"],
    }

    return QuicConfiguration(is_client=is_client, **config_params)


def get_stream_direction_from_id(*, stream_id: StreamId, is_client: bool) -> StreamDirection:
    """Determine the stream direction from its ID and the endpoint role."""
    validate_stream_id(stream_id=stream_id)

    match (
        is_bidirectional_stream(stream_id=stream_id),
        can_send_data_on_stream(stream_id=stream_id, is_client=is_client),
    ):
        case (True, _):
            return StreamDirection.BIDIRECTIONAL
        case (False, True):
            return StreamDirection.SEND_ONLY
        case (False, False):
            return StreamDirection.RECEIVE_ONLY
        case _:
            raise AssertionError("Unreachable code: Invalid stream direction logic")


def is_bidirectional_stream(*, stream_id: StreamId) -> bool:
    """Check if a stream is bidirectional."""
    return (stream_id & 0x2) == 0


def is_client_initiated_stream(*, stream_id: StreamId) -> bool:
    """Check if a stream was initiated by the client (stream IDs are even)."""
    return (stream_id & 0x1) == 0


def is_server_initiated_stream(*, stream_id: StreamId) -> bool:
    """Check if a stream was initiated by the server (stream IDs are odd)."""
    return (stream_id & 0x1) == 1


def is_unidirectional_stream(*, stream_id: StreamId) -> bool:
    """Check if a stream is unidirectional."""
    return (stream_id & 0x2) != 0


def webtransport_code_to_http_code(app_error_code: int) -> int:
    """Maps a 32-bit WebTransport application error code to an HTTP/3 error code."""
    if not (0x0 <= app_error_code <= 0xFFFFFFFF):
        raise ValueError("Application error code must be a 32-bit unsigned integer.")

    return ErrorCodes.WT_APPLICATION_ERROR_FIRST + app_error_code + (app_error_code // 0x1E)



"""WebTransport Protocol Handler."""

from __future__ import annotations

import asyncio
import weakref
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Self

from aioquic.buffer import Buffer, BufferReadError, encode_uint_var
from aioquic.quic.connection import QuicConnection, QuicConnectionState
from aioquic.quic.events import QuicEvent, StreamReset

from pywebtransport import constants
from pywebtransport.config import ClientConfig, ServerConfig
from pywebtransport.constants import ErrorCodes
from pywebtransport.events import Event, EventEmitter
from pywebtransport.exceptions import ConnectionError, FlowControlError, ProtocolError, TimeoutError
from pywebtransport.protocol import utils as protocol_utils
from pywebtransport.protocol.events import (
    CapsuleReceived,
    DatagramReceived,
    H3Event,
    HeadersReceived,
    WebTransportStreamDataReceived,
)
from pywebtransport.protocol.h3_engine import WebTransportH3Engine
from pywebtransport.protocol.session_info import StreamInfo, WebTransportSessionInfo
from pywebtransport.types import (
    ConnectionState,
    EventHandler,
    EventType,
    Headers,
    SessionId,
    SessionState,
    StreamDirection,
    StreamId,
    StreamState,
)
from pywebtransport.utils import Timer, generate_session_id, get_logger, get_timestamp, validate_session_id

if TYPE_CHECKING:
    from pywebtransport.connection import WebTransportConnection


__all__ = ["WebTransportProtocolHandler"]

logger = get_logger(name="protocol.handler")


class WebTransportProtocolHandler(EventEmitter):
    """Orchestrates WebTransport sessions and streams over a QUIC connection."""

    def __init__(
        self,
        *,
        quic_connection: QuicConnection,
        is_client: bool = True,
        connection: WebTransportConnection | None = None,
    ) -> None:
        """Initialize the WebTransport protocol handler."""
        super().__init__()
        self._quic = quic_connection
        self._is_client = is_client
        self._connection_ref = weakref.ref(connection) if connection else None
        self._config = (
            connection.config if connection else (ClientConfig.create() if is_client else ServerConfig.create())
        )

        self._h3: WebTransportH3Engine = WebTransportH3Engine(quic=self._quic, config=self._config)
        self._h3.on(event_type=EventType.SETTINGS_RECEIVED, handler=self._on_settings_received)

        self._capsule_received_handler: EventHandler = self._create_capsule_received_handler()
        self._h3.on(event_type=EventType.CAPSULE_RECEIVED, handler=self._capsule_received_handler)

        self._sessions: dict[SessionId, WebTransportSessionInfo] = {}
        self._streams: dict[StreamId, StreamInfo] = {}
        self._session_control_streams: dict[StreamId, SessionId] = {}
        self._data_stream_to_session: dict[StreamId, SessionId] = {}
        self._session_owned_streams: dict[SessionId, set[StreamId]] = defaultdict(set)
        self._connection_state: ConnectionState = ConnectionState.IDLE
        self._last_activity = get_timestamp()
        self._peer_max_sessions: int | None = None
        self._peer_initial_max_data: int = 0
        self._peer_initial_max_streams_bidi: int = 0
        self._peer_initial_max_streams_uni: int = 0
        self._stats: dict[str, Any] = {
            "bytes_sent": 0,
            "bytes_received": 0,
            "sessions_created": 0,
            "streams_created": 0,
            "datagrams_sent": 0,
            "datagrams_received": 0,
            "errors": 0,
            "connected_at": None,
        }
        self._pending_events: defaultdict[StreamId, list[tuple[float, H3Event]]] = defaultdict(list)
        self._pending_events_count: int = 0
        self._cleanup_pending_events_task: asyncio.Task[None] | None = None
        logger.debug("WebTransport protocol handler initialized (client=%s)", is_client)

    @classmethod
    def create(
        cls,
        *,
        quic_connection: QuicConnection,
        is_client: bool = True,
        connection: WebTransportConnection | None = None,
    ) -> Self:
        """Factory method to create a new WebTransport protocol handler instance."""
        return cls(quic_connection=quic_connection, is_client=is_client, connection=connection)

    @property
    def is_connected(self) -> bool:
        """Check if the underlying connection is established."""
        return self._connection_state == ConnectionState.CONNECTED

    @property
    def connection(self) -> WebTransportConnection | None:
        """Get the parent WebTransportConnection via a weak reference."""
        return self._connection_ref() if self._connection_ref else None

    @property
    def connection_state(self) -> ConnectionState:
        """Get the current state of the underlying connection."""
        return self._connection_state

    @property
    def quic_connection(self) -> QuicConnection:
        """Get the underlying aioquic QuicConnection object."""
        return self._quic

    @property
    def stats(self) -> dict[str, Any]:
        """Get a copy of the protocol handler's statistics."""
        return self._stats.copy()

    async def close(self) -> None:
        """Close the protocol handler and clean up its resources."""
        if self._connection_state == ConnectionState.CLOSED:
            return

        if self._cleanup_pending_events_task:
            self._cleanup_pending_events_task.cancel()
            try:
                await self._cleanup_pending_events_task
            except asyncio.CancelledError:
                pass

        self._connection_state = ConnectionState.CLOSED
        self._teardown_event_handlers()
        await super().close()

    def connection_established(self) -> None:
        """Signal that the QUIC connection is established."""
        if self._connection_state in [ConnectionState.IDLE, ConnectionState.CONNECTING]:
            self._connection_state = ConnectionState.CONNECTED
            self._stats["connected_at"] = get_timestamp()
            if self._config.pending_event_ttl > 0:
                self._cleanup_pending_events_task = asyncio.create_task(self._cleanup_pending_events_loop())
            logger.info("Connection established.")
            self._trigger_transmission()

    def abort_stream(self, *, stream_id: StreamId, error_code: int) -> None:
        """Abort a stream immediately."""
        if stream_id not in self._quic._streams:
            self._cleanup_stream(stream_id=stream_id)
            return

        stream = self._quic._streams[stream_id]
        logger.info("Aborting stream %d with error code 0x%x", stream_id, error_code)

        protocol_error_code = error_code
        if error_code < 2**32:
            try:
                protocol_error_code = protocol_utils.webtransport_code_to_http_code(app_error_code=error_code)
            except ValueError:
                protocol_error_code = ErrorCodes.INTERNAL_ERROR

        can_send = protocol_utils.can_send_data_on_stream(stream_id=stream_id, is_client=self._is_client)
        can_receive = protocol_utils.can_receive_data_on_stream(stream_id=stream_id, is_client=self._is_client)

        try:
            if can_send and stream.sender._reset_error_code is None:
                self._quic.reset_stream(stream_id=stream_id, error_code=protocol_error_code)
            if can_receive and not stream.receiver.is_finished:
                self._quic.stop_stream(stream_id=stream_id, error_code=protocol_error_code)
        except ValueError as e:
            logger.warning("Failed to abort stream %d at QUIC layer: %s", stream_id, e)

        self._trigger_transmission()
        self._cleanup_stream(stream_id=stream_id)

    def accept_webtransport_session(self, *, stream_id: StreamId, session_id: SessionId) -> None:
        """Accept a pending WebTransport session (server-only)."""
        if self._is_client:
            raise ProtocolError(message="Only servers can accept WebTransport sessions")

        session_info = self._sessions.get(session_id)
        if not session_info or session_info.stream_id != stream_id:
            raise ProtocolError(message=f"No pending session found for stream {stream_id} and id {session_id}")

        self._h3.send_headers(stream_id=stream_id, headers={":status": "200"})
        session_info.state = SessionState.CONNECTED
        session_info.ready_at = get_timestamp()
        self._trigger_transmission()

        asyncio.create_task(self.emit(event_type=EventType.SESSION_READY, data=session_info.to_dict()))
        logger.info("Accepted WebTransport session: %s", session_id)

    def close_webtransport_session(self, *, session_id: SessionId, code: int = 0, reason: str | None = None) -> None:
        """Close a specific WebTransport session."""
        session_info = self._sessions.get(session_id)
        if not session_info or session_info.state == SessionState.CLOSED:
            return

        logger.info("Closing WebTransport session: %s (code=%d)", session_id, code)

        buf = Buffer(capacity=1024)
        buf.push_uint32(code)
        buf.push_bytes((reason or "").encode("utf-8"))
        payload = buf.data
        capsule_header = encode_uint_var(constants.CLOSE_WEBTRANSPORT_SESSION_TYPE) + encode_uint_var(len(payload))

        self._h3.send_capsule(
            stream_id=session_info.stream_id,
            capsule_data=capsule_header + payload,
        )
        self._quic.send_stream_data(stream_id=session_info.stream_id, data=b"", end_stream=True)
        self._trigger_transmission()
        self._cleanup_session(session_id=session_id)

        asyncio.create_task(
            self.emit(
                event_type=EventType.SESSION_CLOSED,
                data={"session_id": session_id, "code": code, "reason": reason},
            )
        )

    async def create_webtransport_session(
        self, *, path: str, headers: Headers | None = None
    ) -> tuple[SessionId, StreamId]:
        """Initiate a new WebTransport session (client-only)."""
        if not self._is_client:
            raise ProtocolError(message="Only clients can create WebTransport sessions")

        if self._peer_max_sessions is not None and len(self._sessions) >= self._peer_max_sessions:
            raise ConnectionError(
                message=f"Cannot create new session: server's session limit ({self._peer_max_sessions}) reached."
            )

        session_id = generate_session_id()
        headers_dict = headers or {}
        server_name = self._quic.configuration.server_name
        authority = headers_dict.get("host") or server_name or "localhost"
        connect_headers: Headers = {
            ":method": "CONNECT",
            ":protocol": "webtransport",
            ":scheme": "https",
            ":path": path,
            ":authority": authority,
            **headers_dict,
        }
        stream_id = self._quic.get_next_available_stream_id(is_unidirectional=False)
        self._h3.send_headers(stream_id=stream_id, headers=connect_headers, end_stream=False)

        session_info = WebTransportSessionInfo(
            session_id=session_id,
            stream_id=stream_id,
            state=SessionState.CONNECTING,
            created_at=get_timestamp(),
            path=path,
            headers=headers_dict,
        )
        self._register_session(session_id=session_id, session_info=session_info)
        self._trigger_transmission()
        logger.info(
            "Initiated WebTransport session: %s on control stream %d",
            session_id,
            stream_id,
        )
        return session_id, stream_id

    def create_webtransport_stream(self, *, session_id: SessionId, is_unidirectional: bool = False) -> StreamId:
        """Create a new WebTransport data stream for a session."""
        session_info = self._sessions.get(session_id)
        if not session_info or session_info.state != SessionState.CONNECTED:
            raise ProtocolError(message=f"Session {session_id} not found or not ready")

        if is_unidirectional:
            if session_info.local_streams_uni_opened >= session_info.peer_max_streams_uni:
                self._send_blocked_capsule(session_info=session_info, is_unidirectional=True)
                raise FlowControlError(message="Unidirectional stream limit reached for session.")
            session_info.local_streams_uni_opened += 1
        else:
            if session_info.local_streams_bidi_opened >= session_info.peer_max_streams_bidi:
                self._send_blocked_capsule(session_info=session_info, is_unidirectional=False)
                raise FlowControlError(message="Bidirectional stream limit reached for session.")
            session_info.local_streams_bidi_opened += 1

        stream_id = self._h3.create_webtransport_stream(
            session_id=session_info.stream_id, is_unidirectional=is_unidirectional
        )
        direction = StreamDirection.SEND_ONLY if is_unidirectional else StreamDirection.BIDIRECTIONAL
        self._register_stream(session_id=session_id, stream_id=stream_id, direction=direction)
        self._trigger_transmission()
        logger.debug("Created WebTransport stream %d (%s)", stream_id, direction)
        return stream_id

    async def establish_session(
        self, *, path: str, headers: Headers | None = None, timeout: float = 30.0
    ) -> tuple[SessionId, StreamId]:
        """Establish a WebTransport session with a specified timeout."""
        if not self.is_connected:
            raise ConnectionError(message="Protocol not connected")

        with Timer(name="establish_session") as timer:
            session_id, stream_id = await asyncio.wait_for(
                self.create_webtransport_session(path=path, headers=headers),
                timeout=timeout,
            )

            def session_ready_condition(event: Event) -> bool:
                return isinstance(event.data, dict) and event.data.get("session_id") == session_id

            await self.wait_for(
                event_type=EventType.SESSION_READY,
                timeout=timeout,
                condition=session_ready_condition,
            )
            logger.info(
                "WebTransport session established in %.2fs: %s",
                timer.elapsed,
                session_id,
            )
            return session_id, stream_id

    async def handle_quic_event(self, *, event: QuicEvent) -> None:
        """Process a QUIC event through the H3 engine and handle results."""
        if self._connection_state == ConnectionState.CLOSED:
            return

        if isinstance(event, StreamReset):
            await self._handle_stream_reset(event=event)

        h3_events = await self._h3.handle_event(event=event)
        for h3_event in h3_events:
            await self._handle_h3_event(h3_event=h3_event)

        self._trigger_transmission()

    async def read_stream_complete(self, *, stream_id: StreamId, timeout: float = 30.0) -> bytes:
        """Receive all data from a stream until it is ended."""
        chunks: list[bytes] = []
        future = asyncio.get_running_loop().create_future()

        async def data_handler(event: Event) -> None:
            if future.done():
                return
            if event.data:
                chunks.append(event.data.get("data", b""))
                if event.data.get("end_stream"):
                    future.set_result(None)

        event_name = f"stream_data_received:{stream_id}"
        self.on(event_type=event_name, handler=data_handler)
        try:
            await asyncio.wait_for(future, timeout=timeout)
        except asyncio.TimeoutError:
            raise TimeoutError(message=f"Timeout waiting for stream {stream_id} to end") from None
        finally:
            self.off(event_type=event_name, handler=data_handler)
        return b"".join(chunks)

    def send_webtransport_datagram(self, *, session_id: SessionId, data: bytes) -> None:
        """Send a WebTransport datagram for a session."""
        session_info = self._sessions.get(session_id)
        if not session_info or session_info.state != SessionState.CONNECTED:
            raise ProtocolError(message=f"Session {session_id} not found or not ready")

        self._h3.send_datagram(stream_id=session_info.stream_id, data=data)
        self._stats["bytes_sent"] += len(data)
        self._stats["datagrams_sent"] += 1
        self._trigger_transmission()

    def send_webtransport_stream_data(self, *, stream_id: StreamId, data: bytes, end_stream: bool = False) -> None:
        """Send data on a specific WebTransport stream."""
        stream_info = self._streams.get(stream_id)
        if not stream_info or stream_info.state in (
            StreamState.HALF_CLOSED_LOCAL,
            StreamState.CLOSED,
        ):
            raise ProtocolError(message=f"Stream {stream_id} not found or not writable")

        session_info = self._sessions.get(stream_info.session_id)
        if not session_info:
            raise ProtocolError(message=f"No session found for stream {stream_id}")

        data_len = len(data)
        if session_info.local_data_sent + data_len > session_info.peer_max_data:
            self._send_blocked_capsule(session_info=session_info, is_data=True)
            raise FlowControlError(message="Session data limit reached.")
        session_info.local_data_sent += data_len

        self._h3.send_data(stream_id=stream_id, data=data, end_stream=end_stream)
        self._stats["bytes_sent"] += data_len
        stream_info.bytes_sent += data_len
        self._trigger_transmission()

        if end_stream:
            self._update_stream_state_on_send_end(stream_id=stream_id)

    def get_all_sessions(self) -> list[WebTransportSessionInfo]:
        """Get a list of all current sessions."""
        return list(self._sessions.values())

    def get_health_status(self) -> dict[str, Any]:
        """Get the overall health status of the protocol handler."""
        stats = self.stats
        sessions = self.get_all_sessions()
        streams = list(self._streams.values())
        active_sessions = sum(1 for s in sessions if s.state == SessionState.CONNECTED)
        active_streams = sum(1 for s in streams if s.state == StreamState.OPEN)
        error_rate = stats.get("errors", 0) / max(1, stats.get("sessions_created", 1))

        health_status = "healthy"
        if error_rate > 0.1:
            health_status = "degraded"
        elif not self.is_connected:
            health_status = "unhealthy"

        return {
            "status": health_status,
            "connection_state": self.connection_state,
            "active_sessions": active_sessions,
            "active_streams": active_streams,
            "total_sessions": len(sessions),
            "total_streams": len(streams),
            "error_rate": error_rate,
            "last_activity": stats.get("last_activity"),
            "uptime": (get_timestamp() - stats["connected_at"]) if stats.get("connected_at") else 0.0,
        }

    def get_session_info(self, *, session_id: SessionId) -> WebTransportSessionInfo | None:
        """Get information about a specific session."""
        return self._sessions.get(session_id)

    async def recover_session(self, *, session_id: SessionId, max_retries: int = 3) -> bool:
        """Attempt to recover a failed session by creating a new one."""
        validate_session_id(session_id=session_id)
        session_info = self.get_session_info(session_id=session_id)
        if not session_info:
            return False

        for attempt in range(max_retries):
            try:
                new_session_id, _ = await self.create_webtransport_session(
                    path=session_info.path, headers=session_info.headers
                )
                logger.info(
                    "Recovered session %s as new session %s (attempt %d)",
                    session_id,
                    new_session_id,
                    attempt + 1,
                )
                return True
            except Exception as e:
                logger.warning(
                    "Session recovery attempt %d failed: %s",
                    attempt + 1,
                    e,
                    exc_info=True,
                )
                if attempt < max_retries - 1:
                    await asyncio.sleep(2**attempt)
        return False

    async def _cleanup_pending_events_loop(self) -> None:
        """Periodically clean up stale pending events."""
        try:
            while True:
                now = get_timestamp()
                expired_keys = [
                    key
                    for key, events in self._pending_events.items()
                    if events and (now - events[0][0]) > self._config.pending_event_ttl
                ]

                for session_stream_id in expired_keys:
                    events_to_discard = self._pending_events.pop(session_stream_id, [])
                    self._pending_events_count -= len(events_to_discard)
                    logger.warning(
                        "Discarding %d expired pending events for unknown session stream %d",
                        len(events_to_discard),
                        session_stream_id,
                    )
                    for _, event in events_to_discard:
                        if isinstance(event, WebTransportStreamDataReceived):
                            self.abort_stream(
                                stream_id=event.stream_id, error_code=ErrorCodes.WT_BUFFERED_STREAM_REJECTED
                            )
                await asyncio.sleep(self._config.pending_event_ttl)
        except asyncio.CancelledError:
            pass

    def _cleanup_session(self, *, session_id: SessionId) -> None:
        """Remove a session and all its associated streams."""
        if session_info := self._sessions.pop(session_id, None):
            self._session_control_streams.pop(session_info.stream_id, None)
            stream_ids_to_reset = list(self._session_owned_streams.pop(session_id, set()))
            for stream_id in stream_ids_to_reset:
                if stream_id in self._streams:
                    self.abort_stream(stream_id=stream_id, error_code=ErrorCodes.WT_SESSION_GONE)
            logger.info("Cleaned up session %s and its associated streams.", session_id)

    def _cleanup_stream(self, *, stream_id: StreamId) -> None:
        """Remove a single stream from internal tracking."""
        if self._streams.pop(stream_id, None):
            session_id = self._data_stream_to_session.pop(stream_id, None)
            if session_id and session_id in self._session_owned_streams:
                self._session_owned_streams[session_id].discard(stream_id)
            asyncio.create_task(self.emit(event_type=f"stream_closed:{stream_id}"))

    def _create_capsule_received_handler(self) -> EventHandler:
        """Create a handler for capsule received events."""

        async def capsule_handler(event: Event) -> None:
            if isinstance(event, CapsuleReceived):
                await self._on_capsule_received(event)

        return capsule_handler

    async def _handle_datagram_received(self, *, event: DatagramReceived) -> None:
        """Handle a datagram received from the H3 engine."""
        if connection := self.connection:
            if hasattr(connection, "record_activity"):
                connection.record_activity()
        self._last_activity = get_timestamp()

        session_id = self._session_control_streams.get(event.stream_id)
        if session_id and self._sessions.get(session_id):
            self._stats["bytes_received"] += len(event.data)
            self._stats["datagrams_received"] += 1
            await self.emit(
                event_type=EventType.DATAGRAM_RECEIVED,
                data={"session_id": session_id, "data": event.data},
            )
        elif self._config.pending_event_ttl > 0:
            if self._pending_events_count >= self._config.max_total_pending_events:
                logger.warning("Global pending event buffer full (%d), dropping datagram.", self._pending_events_count)
                return
            if len(self._pending_events.get(event.stream_id, [])) >= self._config.max_pending_events_per_session:
                logger.warning("Pending event buffer full for session stream %d, dropping datagram.", event.stream_id)
                return

            logger.debug("Buffering datagram for unknown session stream %d", event.stream_id)
            self._pending_events[event.stream_id].append((get_timestamp(), event))
            self._pending_events_count += 1

    async def _handle_h3_event(self, *, h3_event: H3Event) -> None:
        """Route H3 events to their specific handlers."""
        match h3_event:
            case HeadersReceived():
                await self._handle_session_headers(event=h3_event)
            case WebTransportStreamDataReceived():
                await self._handle_webtransport_stream_data(event=h3_event)
            case DatagramReceived():
                await self._handle_datagram_received(event=h3_event)
            case CapsuleReceived() as capsule_event:
                await self._on_capsule_received(capsule_event)
            case _:
                logger.debug("Ignoring unhandled H3 event: %s", type(h3_event))

    async def _handle_session_headers(self, *, event: HeadersReceived) -> None:
        """Handle HEADERS frames for session negotiation."""
        if connection := self.connection:
            if hasattr(connection, "record_activity"):
                connection.record_activity()
        self._last_activity = get_timestamp()
        headers_dict = event.headers
        logger.debug("H3 headers received on stream %d: %s", event.stream_id, headers_dict)

        match (self._is_client, headers_dict.get(":method"), headers_dict.get(":protocol")):
            case (True, _, _):
                if session_id := self._session_control_streams.get(event.stream_id):
                    if session_id in self._sessions and headers_dict.get(":status") == "200":
                        session = self._sessions[session_id]
                        session.state = SessionState.CONNECTED
                        session.ready_at = get_timestamp()
                        logger.info("Client session %s is ready.", session_id)
                        await self.emit(event_type=EventType.SESSION_READY, data=session.to_dict())
                        self._process_pending_events(connect_stream_id=event.stream_id)
                    elif session_id:
                        status = headers_dict.get(":status", "unknown")
                        logger.error("Session %s creation failed with status %s", session_id, status)
                        await self.emit(
                            event_type=EventType.SESSION_CLOSED,
                            data={
                                "session_id": session_id,
                                "code": 1,
                                "reason": f"HTTP status {status}",
                            },
                        )
                        self._cleanup_session(session_id=session_id)
            case (False, "CONNECT", "webtransport"):
                my_limit = 1
                if (connection := self.connection) and isinstance(connection.config, ServerConfig):
                    my_limit = connection.config.max_sessions

                if len(self._sessions) >= my_limit:
                    logger.warning(
                        "Session limit (%d) exceeded. Rejecting new session on stream %d",
                        my_limit,
                        event.stream_id,
                    )
                    self._quic.reset_stream(stream_id=event.stream_id, error_code=ErrorCodes.H3_REQUEST_REJECTED)
                    self._trigger_transmission()
                    return

                session_id = generate_session_id()
                app_headers = headers_dict
                session_info = WebTransportSessionInfo(
                    session_id=session_id,
                    stream_id=event.stream_id,
                    state=SessionState.CONNECTING,
                    created_at=get_timestamp(),
                    path=app_headers.get(":path", "/"),
                    headers=app_headers,
                )
                self._register_session(session_id=session_id, session_info=session_info)
                self._process_pending_events(connect_stream_id=event.stream_id)
                event_data = session_info.to_dict()
                if connection := self.connection:
                    event_data["connection"] = connection
                logger.info(
                    "Received WebTransport session request: %s for path '%s'",
                    session_id,
                    session_info.path,
                )
                await self.emit(event_type=EventType.SESSION_REQUEST, data=event_data)
            case (False, method, _):
                logger.warning(
                    "Rejecting unsupported H3 request (method=%s, path=%s) on stream %d. "
                    "This server only accepts WebTransport CONNECT requests.",
                    method,
                    headers_dict.get(":path"),
                    event.stream_id,
                )
                try:
                    self._h3.send_headers(
                        stream_id=event.stream_id,
                        headers={":status": "404"},
                        end_stream=True,
                    )
                    self._trigger_transmission()
                except Exception as e:
                    logger.debug("Failed to send 404 rejection, aborting stream: %s", e)
                    self.abort_stream(stream_id=event.stream_id, error_code=ErrorCodes.H3_REQUEST_REJECTED)

    async def _handle_stream_reset(self, *, event: StreamReset) -> None:
        """Handle a reset stream event."""
        if session_id := self._session_control_streams.get(event.stream_id):
            logger.info(
                "Session %s closed due to control stream %d reset.",
                session_id,
                event.stream_id,
            )
            await self.emit(
                event_type=EventType.SESSION_CLOSED,
                data={
                    "session_id": session_id,
                    "code": event.error_code,
                    "reason": "Control stream reset",
                },
            )
            self._cleanup_session(session_id=session_id)
        else:
            self._cleanup_stream(stream_id=event.stream_id)

    async def _handle_webtransport_stream_data(self, *, event: WebTransportStreamDataReceived) -> None:
        """Handle data received on a WebTransport data stream."""
        if connection := self.connection:
            if hasattr(connection, "record_activity"):
                connection.record_activity()
        stream_id = event.stream_id
        session_stream_id = event.session_id

        session_id = self._session_control_streams.get(session_stream_id)
        if not session_id:
            if self._config.pending_event_ttl > 0:
                if self._pending_events_count >= self._config.max_total_pending_events:
                    logger.warning(
                        "Global pending event buffer full (%d), rejecting stream %d.",
                        self._pending_events_count,
                        stream_id,
                    )
                    self.abort_stream(stream_id=stream_id, error_code=ErrorCodes.WT_BUFFERED_STREAM_REJECTED)
                    return
                if len(self._pending_events.get(session_stream_id, [])) >= self._config.max_pending_events_per_session:
                    logger.warning(
                        "Pending event buffer full for session stream %d, rejecting stream %d.",
                        session_stream_id,
                        stream_id,
                    )
                    self.abort_stream(stream_id=stream_id, error_code=ErrorCodes.WT_BUFFERED_STREAM_REJECTED)
                    return

                logger.debug("Buffering stream data for unknown session stream %d", session_stream_id)
                self._pending_events[session_stream_id].append((get_timestamp(), event))
                self._pending_events_count += 1
            elif self._quic._state not in (
                QuicConnectionState.CLOSING,
                QuicConnectionState.DRAINING,
            ):
                logger.error(
                    "No session mapping found for session_stream_id %d on new stream %d.",
                    session_stream_id,
                    stream_id,
                )
            return

        session_info = self._sessions[session_id]
        session_info.peer_data_sent += len(event.data)

        if stream_id not in self._data_stream_to_session:
            direction = protocol_utils.get_stream_direction_from_id(stream_id=stream_id, is_client=self._is_client)
            self._register_stream(session_id=session_id, stream_id=stream_id, direction=direction)

            if protocol_utils.is_unidirectional_stream(stream_id=stream_id):
                session_info.peer_streams_uni_opened += 1
            else:
                session_info.peer_streams_bidi_opened += 1

            event_data = self._streams[stream_id].to_dict()
            event_data["initial_payload"] = {
                "data": event.data,
                "end_stream": event.stream_ended,
            }
            await self.emit(event_type=EventType.STREAM_OPENED, data=event_data)
        else:
            await self.emit(
                event_type=f"stream_data_received:{stream_id}",
                data={"data": event.data, "end_stream": event.stream_ended},
            )

        await self._update_local_flow_control(session_id=session_id)

    async def _on_capsule_received(self, event: CapsuleReceived) -> None:
        """Handle a CAPSULE_RECEIVED event from the H3 engine."""
        session_id = self._session_control_streams.get(event.stream_id)
        if not (session_id and (session_info := self._sessions.get(session_id))):
            return

        try:
            raw_data = event.capsule_data
            buf = Buffer(data=raw_data)

            match event.capsule_type:
                case constants.WT_MAX_DATA_TYPE:
                    new_limit = buf.pull_uint_var()
                    if new_limit > session_info.peer_max_data:
                        session_info.peer_max_data = new_limit
                        await self.emit(
                            event_type=EventType.SESSION_MAX_DATA_UPDATED,
                            data={"session_id": session_id, "max_data": new_limit},
                        )
                    elif new_limit < session_info.peer_max_data:
                        raise ProtocolError(message="Flow control limit decreased for MAX_DATA")
                case constants.WT_MAX_STREAMS_BIDI_TYPE:
                    new_limit = buf.pull_uint_var()
                    if new_limit > session_info.peer_max_streams_bidi:
                        session_info.peer_max_streams_bidi = new_limit
                        await self.emit(
                            event_type=EventType.SESSION_MAX_STREAMS_BIDI_UPDATED,
                            data={
                                "session_id": session_id,
                                "max_streams_bidi": new_limit,
                            },
                        )
                    elif new_limit < session_info.peer_max_streams_bidi:
                        raise ProtocolError(message="Flow control limit decreased for MAX_STREAMS_BIDI")
                case constants.WT_MAX_STREAMS_UNI_TYPE:
                    new_limit = buf.pull_uint_var()
                    if new_limit > session_info.peer_max_streams_uni:
                        session_info.peer_max_streams_uni = new_limit
                        await self.emit(
                            event_type=EventType.SESSION_MAX_STREAMS_UNI_UPDATED,
                            data={
                                "session_id": session_id,
                                "max_streams_uni": new_limit,
                            },
                        )
                    elif new_limit < session_info.peer_max_streams_uni:
                        raise ProtocolError(message="Flow control limit decreased for MAX_STREAMS_UNI")
                case constants.CLOSE_WEBTRANSPORT_SESSION_TYPE:
                    app_code = buf.pull_uint32()
                    reason_bytes = buf.pull_bytes(len(raw_data) - buf.tell())
                    try:
                        reason = reason_bytes.decode("utf-8")
                    except UnicodeDecodeError:
                        logger.warning(
                            "Received CLOSE_SESSION capsule for session %s with invalid UTF-8 reason string.",
                            session_id,
                        )
                        reason = reason_bytes.decode("utf-8", errors="replace")
                    logger.info(
                        "Received CLOSE_SESSION capsule: code=%d reason=%s",
                        app_code,
                        reason,
                    )
                    await self.emit(
                        event_type=EventType.SESSION_CLOSED,
                        data={
                            "session_id": session_id,
                            "code": app_code,
                            "reason": reason,
                        },
                    )
                    self._cleanup_session(session_id=session_id)

                case constants.DRAIN_WEBTRANSPORT_SESSION_TYPE:
                    logger.info("Received DRAIN_SESSION capsule for session %s", session_id)
                    await self.emit(
                        event_type=EventType.SESSION_DRAINING,
                        data={"session_id": session_id},
                    )

        except BufferReadError:
            logger.warning("Could not parse flow control capsule for session %s", session_id)

    async def _on_settings_received(self, event: Event) -> None:
        """Handle the SETTINGS_RECEIVED event from the H3 engine."""
        if isinstance(event.data, dict) and (settings := event.data.get("settings")):
            self._peer_max_sessions = settings.get(constants.SETTINGS_WT_MAX_SESSIONS)
            self._peer_initial_max_data = settings.get(constants.SETTINGS_WT_INITIAL_MAX_DATA, 0)
            self._peer_initial_max_streams_bidi = settings.get(constants.SETTINGS_WT_INITIAL_MAX_STREAMS_BIDI, 0)
            self._peer_initial_max_streams_uni = settings.get(constants.SETTINGS_WT_INITIAL_MAX_STREAMS_UNI, 0)

            for session in self._sessions.values():
                session.peer_max_data = self._peer_initial_max_data
                session.peer_max_streams_bidi = self._peer_initial_max_streams_bidi
                session.peer_max_streams_uni = self._peer_initial_max_streams_uni

    async def _process_buffered_events(self, events: list[tuple[float, H3Event]]) -> None:
        """Asynchronously process a list of buffered events."""
        for _, event in events:
            if isinstance(event, WebTransportStreamDataReceived):
                await self._handle_webtransport_stream_data(event=event)
            elif isinstance(event, DatagramReceived):
                await self._handle_datagram_received(event=event)

    def _process_pending_events(self, *, connect_stream_id: StreamId) -> None:
        """Check for and process any buffered events for a newly established session."""
        if events_to_process := self._pending_events.pop(connect_stream_id, None):
            self._pending_events_count -= len(events_to_process)
            logger.debug(
                "Processing %d buffered events for session stream %d", len(events_to_process), connect_stream_id
            )
            asyncio.create_task(self._process_buffered_events(events_to_process))

    def _register_session(self, *, session_id: SessionId, session_info: WebTransportSessionInfo) -> None:
        """Add a new session to internal tracking."""
        self._sessions[session_id] = session_info
        self._session_control_streams[session_info.stream_id] = session_id
        session_info.local_max_data = self._config.initial_max_data
        session_info.local_max_streams_bidi = self._config.initial_max_streams_bidi
        session_info.local_max_streams_uni = self._config.initial_max_streams_uni
        session_info.peer_max_data = self._peer_initial_max_data
        session_info.peer_max_streams_bidi = self._peer_initial_max_streams_bidi
        session_info.peer_max_streams_uni = self._peer_initial_max_streams_uni
        self._stats["sessions_created"] += 1

    def _register_stream(self, *, session_id: SessionId, stream_id: StreamId, direction: StreamDirection) -> StreamInfo:
        """Add a new stream to internal tracking."""
        stream_info = StreamInfo(
            stream_id=stream_id,
            session_id=session_id,
            direction=direction,
            state=StreamState.OPEN,
            created_at=get_timestamp(),
        )
        self._streams[stream_id] = stream_info
        self._data_stream_to_session[stream_id] = session_id
        self._session_owned_streams[session_id].add(stream_id)
        self._stats["streams_created"] += 1
        logger.debug("Registered %s stream %d for session %s", direction, stream_id, session_id)
        return stream_info

    def _send_blocked_capsule(
        self,
        *,
        session_info: WebTransportSessionInfo,
        is_data: bool = False,
        is_unidirectional: bool = False,
    ) -> None:
        """Send a WT_DATA_BLOCKED or WT_STREAMS_BLOCKED capsule."""
        if is_data:
            capsule_type = constants.WT_DATA_BLOCKED_TYPE
            limit = session_info.peer_max_data
        elif is_unidirectional:
            capsule_type = constants.WT_STREAMS_BLOCKED_UNI_TYPE
            limit = session_info.peer_max_streams_uni
        else:
            capsule_type = constants.WT_STREAMS_BLOCKED_BIDI_TYPE
            limit = session_info.peer_max_streams_bidi

        payload = encode_uint_var(limit)
        capsule_header = encode_uint_var(capsule_type) + encode_uint_var(len(payload))
        self._h3.send_capsule(
            stream_id=session_info.stream_id,
            capsule_data=capsule_header + payload,
        )

    def _send_max_capsule(self, *, session_info: WebTransportSessionInfo, capsule_type: int, value: int) -> None:
        """Send a WT_MAX_DATA or WT_MAX_STREAMS capsule."""
        payload = encode_uint_var(value)
        capsule_header = encode_uint_var(capsule_type) + encode_uint_var(len(payload))
        self._h3.send_capsule(
            stream_id=session_info.stream_id,
            capsule_data=capsule_header + payload,
        )

    def _teardown_event_handlers(self) -> None:
        """Remove all event listeners to prevent memory leaks."""
        if self._h3:
            self._h3.off(event_type=EventType.SETTINGS_RECEIVED, handler=self._on_settings_received)
            self._h3.off(event_type=EventType.CAPSULE_RECEIVED, handler=self._capsule_received_handler)

    def _trigger_transmission(self) -> None:
        """Trigger the underlying QUIC connection to send pending data."""
        if connection := self.connection:
            if hasattr(connection, "_transmit"):
                connection._transmit()

    async def _update_local_flow_control(self, *, session_id: SessionId) -> None:
        """Check and send flow control updates for the local peer."""
        session_info = self._sessions.get(session_id)
        if not session_info:
            return

        if self._config.flow_control_window_size > 0:
            remaining_credit = session_info.local_max_data - session_info.peer_data_sent
            if remaining_credit < (self._config.flow_control_window_size / 2):
                if self._config.flow_control_window_auto_scale:
                    new_limit = session_info.local_max_data * 2
                else:
                    new_limit = session_info.peer_data_sent + self._config.flow_control_window_size
                if new_limit > session_info.local_max_data:
                    session_info.local_max_data = new_limit
                    self._send_max_capsule(
                        session_info=session_info,
                        capsule_type=constants.WT_MAX_DATA_TYPE,
                        value=new_limit,
                    )
                    self._trigger_transmission()

        if session_info.peer_streams_bidi_opened >= session_info.local_max_streams_bidi:
            new_limit = session_info.local_max_streams_bidi + self._config.stream_flow_control_increment_bidi
            session_info.local_max_streams_bidi = new_limit
            self._send_max_capsule(
                session_info=session_info,
                capsule_type=constants.WT_MAX_STREAMS_BIDI_TYPE,
                value=new_limit,
            )
            self._trigger_transmission()

        if session_info.peer_streams_uni_opened >= session_info.local_max_streams_uni:
            new_limit = session_info.local_max_streams_uni + self._config.stream_flow_control_increment_uni
            session_info.local_max_streams_uni = new_limit
            self._send_max_capsule(
                session_info=session_info,
                capsule_type=constants.WT_MAX_STREAMS_UNI_TYPE,
                value=new_limit,
            )
            self._trigger_transmission()

    def _update_stream_state_on_receive_end(self, *, stream_id: StreamId) -> None:
        """Update stream state when its receiving side is closed."""
        if not (stream_info := self._streams.get(stream_id)):
            return

        new_state = StreamState.HALF_CLOSED_REMOTE
        if stream_info.state == StreamState.HALF_CLOSED_LOCAL:
            new_state = StreamState.CLOSED
        stream_info.state = new_state
        if new_state == StreamState.CLOSED:
            self._cleanup_stream(stream_id=stream_id)

    def _update_stream_state_on_send_end(self, *, stream_id: StreamId) -> None:
        """Update stream state when its sending side is closed."""
        if not (stream_info := self._streams.get(stream_id)):
            return

        new_state = StreamState.HALF_CLOSED_LOCAL
        if stream_info.state == StreamState.HALF_CLOSED_REMOTE:
            new_state = StreamState.CLOSED
        stream_info.state = new_state
        if new_state == StreamState.CLOSED:
            self._cleanup_stream(stream_id=stream_id)



"""WebTransport connection manager implementation."""

from __future__ import annotations

import asyncio
import weakref
from collections import defaultdict
from types import TracebackType
from typing import Any, Self

from pywebtransport.connection.connection import WebTransportConnection
from pywebtransport.constants import (
    DEFAULT_CONNECTION_CLEANUP_INTERVAL,
    DEFAULT_CONNECTION_IDLE_CHECK_INTERVAL,
    DEFAULT_CONNECTION_IDLE_TIMEOUT,
    DEFAULT_SERVER_MAX_CONNECTIONS,
)
from pywebtransport.exceptions import ConnectionError
from pywebtransport.types import ConnectionId, EventType
from pywebtransport.utils import get_logger, get_timestamp

__all__ = ["ConnectionManager"]

logger = get_logger(name="connection.manager")


class ConnectionManager:
    """Manages multiple WebTransport connections with concurrency safety."""

    def __init__(
        self,
        *,
        max_connections: int = DEFAULT_SERVER_MAX_CONNECTIONS,
        connection_cleanup_interval: float = DEFAULT_CONNECTION_CLEANUP_INTERVAL,
        connection_idle_check_interval: float = DEFAULT_CONNECTION_IDLE_CHECK_INTERVAL,
        connection_idle_timeout: float = DEFAULT_CONNECTION_IDLE_TIMEOUT,
    ) -> None:
        """Initialize the connection manager."""
        self._max_connections = max_connections
        self._cleanup_interval = connection_cleanup_interval
        self._idle_check_interval = connection_idle_check_interval
        self._idle_timeout = connection_idle_timeout
        self._lock: asyncio.Lock | None = None
        self._connections: dict[ConnectionId, WebTransportConnection] = {}
        self._stats = {
            "total_created": 0,
            "total_closed": 0,
            "current_count": 0,
            "max_concurrent": 0,
        }
        self._cleanup_task: asyncio.Task[None] | None = None
        self._idle_check_task: asyncio.Task[None] | None = None
        self._is_shutting_down = False

    @classmethod
    def create(
        cls,
        *,
        max_connections: int = DEFAULT_SERVER_MAX_CONNECTIONS,
        connection_cleanup_interval: float = DEFAULT_CONNECTION_CLEANUP_INTERVAL,
        connection_idle_check_interval: float = DEFAULT_CONNECTION_IDLE_CHECK_INTERVAL,
        connection_idle_timeout: float = DEFAULT_CONNECTION_IDLE_TIMEOUT,
    ) -> Self:
        """Factory method to create a new connection manager instance."""
        return cls(
            max_connections=max_connections,
            connection_cleanup_interval=connection_cleanup_interval,
            connection_idle_check_interval=connection_idle_check_interval,
            connection_idle_timeout=connection_idle_timeout,
        )

    async def __aenter__(self) -> Self:
        """Enter async context, initializing resources and starting background tasks."""
        self._lock = asyncio.Lock()
        self._start_background_tasks()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit async context, shutting down the manager."""
        await self.shutdown()

    async def shutdown(self) -> None:
        """Shut down the manager and all associated tasks and connections."""
        if self._is_shutting_down:
            return

        self._is_shutting_down = True

        if self._cleanup_task:
            self._cleanup_task.cancel()
            try:
                await self._cleanup_task
            except asyncio.CancelledError:
                pass

        if self._idle_check_task:
            self._idle_check_task.cancel()
            try:
                await self._idle_check_task
            except asyncio.CancelledError:
                pass

        await self.close_all_connections()
        logger.info("Connection manager shutdown complete")

    async def add_connection(self, *, connection: WebTransportConnection) -> ConnectionId:
        """Add a new connection to the manager."""
        if self._lock is None:
            raise ConnectionError(
                message=(
                    "ConnectionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            if len(self._connections) >= self._max_connections:
                raise ConnectionError(message=f"Maximum connections ({self._max_connections}) exceeded")

            connection_id = connection.connection_id
            self._connections[connection_id] = connection
            manager_ref = weakref.ref(self)

            async def on_close(event: Any) -> None:
                manager = manager_ref()
                if manager and isinstance(event.data, dict):
                    await manager.remove_connection(connection_id=event.data["connection_id"])

            connection.once(event_type=EventType.CONNECTION_CLOSED, handler=on_close)

            self._stats["total_created"] += 1
            self._update_stats_unsafe()
            logger.debug("Added connection %s (total: %d)", connection_id, len(self._connections))
            return connection_id

    async def close_all_connections(self) -> None:
        """Close all currently managed connections."""
        if self._lock is None:
            raise ConnectionError(
                message=(
                    "ConnectionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        connections_to_close: list[WebTransportConnection] = []
        async with self._lock:
            if not self._connections:
                return
            connections_to_close = list(self._connections.values())
            logger.info("Closing %d connections", len(connections_to_close))
            self._connections.clear()

        try:
            async with asyncio.TaskGroup() as tg:
                for conn in connections_to_close:
                    tg.create_task(conn.close())
        except* Exception as eg:
            logger.error("Errors occurred while closing connections: %s", eg.exceptions, exc_info=eg)

        async with self._lock:
            self._stats["total_closed"] += len(connections_to_close)
            self._update_stats_unsafe()

        logger.info("All connections closed")

    async def get_connection(self, *, connection_id: ConnectionId) -> WebTransportConnection | None:
        """Retrieve a connection by its ID."""
        if self._lock is None:
            raise ConnectionError(
                message=(
                    "ConnectionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            return self._connections.get(connection_id)

    async def remove_connection(self, *, connection_id: ConnectionId) -> WebTransportConnection | None:
        """Remove a connection from the manager by its ID."""
        if self._lock is None:
            raise ConnectionError(
                message=(
                    "ConnectionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            connection = self._connections.pop(connection_id, None)
            if connection:
                self._stats["total_closed"] += 1
                self._update_stats_unsafe()
                logger.debug("Removed connection %s (total: %d)", connection_id, len(self._connections))
            return connection

    async def cleanup_closed_connections(self) -> int:
        """Find and remove any connections that are marked as closed."""
        if self._lock is None:
            raise ConnectionError(
                message=(
                    "ConnectionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        closed_connection_ids = []
        async with self._lock:
            for conn_id, conn in list(self._connections.items()):
                if conn.is_closed:
                    closed_connection_ids.append(conn_id)

            for conn_id in closed_connection_ids:
                del self._connections[conn_id]

            if closed_connection_ids:
                self._stats["total_closed"] += len(closed_connection_ids)
                self._update_stats_unsafe()
                logger.debug("Cleaned up %d closed connections.", len(closed_connection_ids))

        return len(closed_connection_ids)

    async def get_all_connections(self) -> list[WebTransportConnection]:
        """Retrieve a list of all current connections."""
        if self._lock is None:
            raise ConnectionError(
                message=(
                    "ConnectionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            return list(self._connections.values())

    def get_connection_count(self) -> int:
        """Get the current number of active connections."""
        return len(self._connections)

    async def get_stats(self) -> dict[str, Any]:
        """Get detailed statistics about the managed connections."""
        if self._lock is None:
            raise ConnectionError(
                message=(
                    "ConnectionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            states: dict[str, int] = defaultdict(int)
            for conn in self._connections.values():
                states[conn.state] += 1
            return {
                **self._stats,
                "active": len(self._connections),
                "states": dict(states),
                "max_connections": self._max_connections,
            }

    def _on_cleanup_done(self, task: asyncio.Task[None]) -> None:
        """Callback to ensure cleanup is triggered when the cleanup task finishes."""
        if self._is_shutting_down:
            return

        if not task.cancelled():
            if exc := task.exception():
                logger.error("Connection cleanup task finished unexpectedly: %s.", exc, exc_info=exc)

        asyncio.create_task(self.shutdown())

    def _on_idle_check_done(self, task: asyncio.Task[None]) -> None:
        """Callback to ensure cleanup is triggered when the idle check task finishes."""
        if self._is_shutting_down:
            return

        if not task.cancelled():
            if exc := task.exception():
                logger.error("Connection idle check task finished unexpectedly: %s.", exc, exc_info=exc)

        asyncio.create_task(self.shutdown())

    async def _periodic_cleanup(self) -> None:
        """Periodically run the cleanup process to remove closed connections."""
        try:
            while True:
                try:
                    await self.cleanup_closed_connections()
                except Exception as e:
                    logger.error("Connection cleanup cycle failed: %s", e, exc_info=e)

                await asyncio.sleep(self._cleanup_interval)
        except (Exception, asyncio.CancelledError):
            pass

    async def _periodic_idle_check(self) -> None:
        """Periodically check for and close idle connections."""
        if self._lock is None:
            logger.warning("Idle check task running on uninitialized ConnectionManager; stopping.")
            return

        try:
            while True:
                try:
                    idle_connections_to_close = []
                    now = get_timestamp()

                    async with self._lock:
                        all_connections = list(self._connections.values())

                    for conn in all_connections:
                        if conn.is_closing or conn.is_closed:
                            continue

                        if self._idle_timeout > 0 and conn.last_activity_time > 0:
                            idle_duration = now - conn.last_activity_time
                            if idle_duration > self._idle_timeout:
                                idle_connections_to_close.append(conn)

                    if idle_connections_to_close:
                        logger.info("Closing %d idle connections.", len(idle_connections_to_close))
                        try:
                            async with asyncio.TaskGroup() as tg:
                                for conn in idle_connections_to_close:
                                    tg.create_task(conn.close(reason="Idle timeout"))
                        except* Exception as eg:
                            logger.error(
                                "Errors occurred while closing idle connections: %s", eg.exceptions, exc_info=eg
                            )

                except Exception as e:
                    logger.error("Idle connection check cycle failed: %s", e, exc_info=e)

                await asyncio.sleep(self._idle_check_interval)
        except (Exception, asyncio.CancelledError):
            pass

    def _start_background_tasks(self) -> None:
        """Start all periodic background tasks."""
        if self._cleanup_task is None or self._cleanup_task.done():
            self._cleanup_task = asyncio.create_task(self._periodic_cleanup())
            self._cleanup_task.add_done_callback(self._on_cleanup_done)

        if self._idle_check_task is None or self._idle_check_task.done():
            self._idle_check_task = asyncio.create_task(self._periodic_idle_check())
            self._idle_check_task.add_done_callback(self._on_idle_check_done)

    def _update_stats_unsafe(self) -> None:
        """Update internal statistics (must be called within a lock)."""
        current_count = len(self._connections)
        self._stats["current_count"] = current_count
        self._stats["max_concurrent"] = max(self._stats["max_concurrent"], current_count)



"""WebTransport Server Cluster."""

from __future__ import annotations

import asyncio
from types import TracebackType
from typing import Any, Self

from pywebtransport.config import ServerConfig
from pywebtransport.exceptions import ServerError
from pywebtransport.server.server import WebTransportServer
from pywebtransport.utils import get_logger

__all__ = ["ServerCluster"]

logger = get_logger(name="server.cluster")


class ServerCluster:
    """Manages the lifecycle of multiple WebTransport server instances."""

    def __init__(self, *, configs: list[ServerConfig]) -> None:
        """Initialize the server cluster."""
        self._configs = configs
        self._servers: list[WebTransportServer] = []
        self._running = False
        self._lock: asyncio.Lock | None = None

    @property
    def is_running(self) -> bool:
        """Check if the cluster is currently running."""
        return self._running

    async def __aenter__(self) -> Self:
        """Enter the async context, initializing resources and starting all servers."""
        self._lock = asyncio.Lock()
        await self.start_all()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the async context and stop all servers."""
        await self.stop_all()

    async def start_all(self) -> None:
        """Start all servers in the cluster concurrently."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            if self._running:
                return
            initial_configs = self._configs

        started_servers: list[WebTransportServer] = []
        tasks: list[asyncio.Task[WebTransportServer]] = []
        try:
            async with asyncio.TaskGroup() as tg:
                for config in initial_configs:
                    tasks.append(tg.create_task(self._create_and_start_server(config=config)))
            started_servers = [task.result() for task in tasks]
        except* Exception as eg:
            logger.error("Failed to start server cluster: %s", eg.exceptions, exc_info=True)
            successful_servers = [task.result() for task in tasks if task.done() and not task.exception()]
            if successful_servers:
                logger.info(
                    "Cleaning up %d successfully started servers...",
                    len(successful_servers),
                )
                try:
                    async with asyncio.TaskGroup() as cleanup_tg:
                        for server in successful_servers:
                            cleanup_tg.create_task(server.close())
                except* Exception as cleanup_eg:
                    logger.error(
                        "Errors during cluster startup cleanup: %s",
                        cleanup_eg.exceptions,
                        exc_info=True,
                    )
            raise eg.exceptions[0]

        async with self._lock:
            self._servers = started_servers
            self._running = True
            logger.info("Started cluster with %d servers", len(self._servers))

    async def stop_all(self) -> None:
        """Stop all servers in the cluster concurrently."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        servers_to_stop: list[WebTransportServer] = []
        async with self._lock:
            if not self._running:
                return
            servers_to_stop = self._servers
            self._servers = []
            self._running = False

        if servers_to_stop:
            try:
                async with asyncio.TaskGroup() as tg:
                    for server in servers_to_stop:
                        tg.create_task(server.close())
            except* Exception as eg:
                logger.error(
                    "Errors occurred while stopping server cluster: %s",
                    eg.exceptions,
                    exc_info=True,
                )
                raise eg
            logger.info("Stopped server cluster")

    async def add_server(self, *, config: ServerConfig) -> WebTransportServer | None:
        """Add and start a new server in the running cluster."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        is_running: bool
        async with self._lock:
            is_running = self._running
            if not is_running:
                self._configs.append(config)
                logger.info("Cluster not running. Server config added for next start.")
                return None

        try:
            server = await self._create_and_start_server(config=config)
            async with self._lock:
                if not self._running:
                    await server.close()
                    logger.warning("Cluster was stopped while new server was starting. New server has been shut down.")
                    return None
                self._servers.append(server)
                logger.info("Added server to cluster: %s", server.local_address)
            return server
        except Exception as e:
            logger.error("Failed to add server to cluster: %s", e, exc_info=True)
            return None

    async def remove_server(self, *, host: str, port: int) -> bool:
        """Remove and stop a specific server from the cluster by its address."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        server_to_remove: WebTransportServer | None = None
        async with self._lock:
            for server in self._servers:
                if server.local_address == (host, port):
                    server_to_remove = server
                    break

            if server_to_remove:
                self._servers.remove(server_to_remove)
            else:
                logger.warning("Server at %s:%s not found in cluster.", host, port)
                return False

        await server_to_remove.close()
        logger.info("Removed server from cluster: %s:%s", host, port)
        return True

    async def get_cluster_stats(self) -> dict[str, Any]:
        """Get deeply aggregated statistics for the entire cluster."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        servers_snapshot: list[WebTransportServer]
        async with self._lock:
            if not self._servers:
                return {}
            servers_snapshot = self._servers.copy()

        tasks: list[asyncio.Task[dict[str, Any]]] = []
        try:
            async with asyncio.TaskGroup() as tg:
                for s in servers_snapshot:
                    tasks.append(tg.create_task(s.get_server_stats()))
        except* Exception as eg:
            logger.error(
                "Failed to fetch stats from some servers: %s",
                eg.exceptions,
                exc_info=True,
            )
            raise eg

        stats_list = [task.result() for task in tasks if task.done() and not task.exception()]

        agg_stats: dict[str, Any] = {
            "server_count": len(servers_snapshot),
            "total_connections_accepted": 0,
            "total_connections_rejected": 0,
            "total_connections_active": 0,
            "total_sessions_active": 0,
        }
        for stats in stats_list:
            agg_stats["total_connections_accepted"] += stats.get("connections_accepted", 0)
            agg_stats["total_connections_rejected"] += stats.get("connections_rejected", 0)
            if "connections" in stats:
                agg_stats["total_connections_active"] += stats["connections"].get("active", 0)
            if "sessions" in stats:
                agg_stats["total_sessions_active"] += stats["sessions"].get("active", 0)

        return agg_stats

    async def get_server_count(self) -> int:
        """Get the number of running servers in the cluster."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )
        async with self._lock:
            return len(self._servers)

    async def get_servers(self) -> list[WebTransportServer]:
        """Get a thread-safe copy of all active servers in the cluster."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )
        async with self._lock:
            return self._servers.copy()

    async def _create_and_start_server(self, *, config: ServerConfig) -> WebTransportServer:
        """Create, activate, and start a single server instance."""
        server = WebTransportServer(config=config)
        await server.__aenter__()
        try:
            await server.listen()
        except Exception:
            await server.close()
            raise
        return server
